import datetime
import random
from os import environ
from pathlib import Path

import numpy as np
import pytest

from act.io.arm import read_arm_netcdf
from act.qc.qcfilter import set_bit
from act.tests import EXAMPLE_MET1
from act.utils.data_utils import DatastreamParserARM


def test_qc_summary():
    for cleanup in [False, True]:
        ds = read_arm_netcdf(EXAMPLE_MET1, cleanup_qc=not cleanup)
        for var_name in ['temp_mean', 'rh_mean']:
            qc_var_name = f'qc_{var_name}'
            qc_data = ds[qc_var_name].values

            assert np.sum(qc_data) == 0

            index_4 = np.arange(100, 200)
            qc_data[index_4] = set_bit(qc_data[index_4], 4)
            index_1 = np.arange(170, 230)
            qc_data[index_1] = set_bit(qc_data[index_1], 1)
            index_2 = np.arange(250, 400)
            qc_data[index_2] = set_bit(qc_data[index_2], 2)
            index_3 = np.arange(450, 510)
            qc_data[index_3] = set_bit(qc_data[index_3], 3)
            ds[qc_var_name].values = qc_data

        result = ds.qcfilter.create_qc_summary(cleanup_qc=cleanup)

        for var_name in ['temp_mean', 'rh_mean']:
            assert 'flag_masks' not in result[qc_var_name].attrs.keys()
            assert isinstance(result[qc_var_name].attrs['flag_values'], list)

            assert np.sum(result[qc_var_name].values) == 880

            qc_ma = result.qcfilter.get_masked_data(var_name, rm_assessments='Suspect')
            assert np.sum(np.where(qc_ma.mask)) == 9415

            qc_ma = result.qcfilter.get_masked_data(var_name, rm_assessments='Incorrect')
            assert np.sum(np.where(qc_ma.mask)) == 89415

            att_names = [
                'fail_min',
                'fail_max',
                'fail_delta',
                'valid_min',
                'valid_max',
                'valid_delta',
            ]
            for att_name in att_names:
                assert att_name not in ds[f'qc_{var_name}'].attrs

        assert "Quality control summary implemented by ACT" in result.attrs['history']

        del ds


def test_qc_summary_multiple_assessment_names():
    ds = read_arm_netcdf(EXAMPLE_MET1, cleanup_qc=True)
    var_name = 'temp_mean'
    qc_var_name = f'qc_{var_name}'
    qc_data = ds[qc_var_name].values

    assert np.sum(qc_data) == 0

    index_4 = np.arange(200, 300)
    qc_data[index_4] = set_bit(qc_data[index_4], 4)
    index_1 = np.arange(270, 330)
    qc_data[index_1] = set_bit(qc_data[index_1], 1)
    index_2 = np.arange(350, 500)
    qc_data[index_2] = set_bit(qc_data[index_2], 2)
    index_3 = np.arange(550, 610)
    qc_data[index_3] = set_bit(qc_data[index_3], 3)
    ds[qc_var_name].values = qc_data

    index_5 = np.arange(50, 150)
    ds.qcfilter.add_test(
        var_name, index=index_5, test_meaning='Testing Suspect', test_assessment='Suspect'
    )

    index_6 = np.arange(130, 210)
    ds.qcfilter.add_test(
        var_name, index=index_6, test_meaning='Testing Incorrect', test_assessment='Incorrect'
    )

    result = ds.qcfilter.create_qc_summary(normalize_assessment=False)

    assert result[qc_var_name].attrs['flag_assessments'] == [
        'Not failing',
        'Suspect',
        'Indeterminate',
        'Incorrect',
        'Bad',
    ]

    qc_ma = result.qcfilter.get_masked_data(var_name, rm_assessments='Indeterminate')
    assert np.sum(np.where(qc_ma.mask)[0]) == 14370

    qc_ma = result.qcfilter.get_masked_data(var_name, rm_assessments='Suspect')
    assert np.sum(np.where(qc_ma.mask)[0]) == 7160

    qc_ma = result.qcfilter.get_masked_data(var_name, rm_assessments='Bad')
    assert np.sum(np.where(qc_ma.mask)[0]) == 116415

    qc_ma = result.qcfilter.get_masked_data(var_name, rm_assessments='Incorrect')
    assert np.sum(np.where(qc_ma.mask)[0]) == 13560

    assert np.sum(np.where(result[qc_var_name].values == 0)) == 884575
    qc_ma = result.qcfilter.get_masked_data(var_name, rm_assessments='Not failing')
    assert np.sum(np.where(qc_ma.mask)[0]) == 884575


def test_qc_summary_unexpected_assessment_name():
    var_name = 'temp_mean'
    ds = read_arm_netcdf(EXAMPLE_MET1, keep_variables=var_name)

    test_meanings = [
        'Testing Bad',
        'Testing Boomer',
        'Testing Boomer Second',
        'Testing Incorrect',
        'Testing Indeterminate',
        'Testing Sooner',
        'Testing Suspect',
    ]
    test_assessments = [
        'Bad',
        'Boomer',
        'boomer',
        'Incorrect',
        'Indeterminate',
        'Sooner',
        'Suspect',
    ]

    test_index_sums = [4950, 39900, 39900, 34950, 44950, 54950, 64950]

    for ii, _ in enumerate(test_assessments):
        ds.qcfilter.add_test(
            var_name,
            index=np.arange(ii * 100, ii * 100 + 100),
            test_meaning=test_meanings[ii],
            test_assessment=test_assessments[ii],
        )

    ds = ds.qcfilter.create_qc_summary(normalize_assessment=False)

    qc_var_name = ds.qcfilter.check_for_ancillary_qc(var_name, add_if_missing=False)

    # Make sure flag meanings are correct with new assessments.
    assert sorted(ds[qc_var_name].attrs['flag_meanings']) == [
        'Data Boomer',
        'Data Sooner',
        'Data incorrect use not recommended',
        'Data incorrect use not recommended',
        'Data suspect further analysis recommended',
        'Data suspect further analysis recommended',
        'Not failing quality control tests',
    ]
    assert sorted(ds[qc_var_name].attrs['flag_assessments']) == [
        'Bad',
        'Boomer',
        'Incorrect',
        'Indeterminate',
        'Not failing',
        'Sooner',
        'Suspect',
    ]
    # Make sure the values and order of first 5 are as expected. The other non-standard
    # assessments may be in different order with set operations.
    assert ds[qc_var_name].attrs['flag_assessments'][:5] == [
        'Not failing',
        'Suspect',
        'Indeterminate',
        'Incorrect',
        'Bad',
    ]

    for assessment, index_sum in zip(test_assessments, test_index_sums):
        qc_ma = ds.qcfilter.get_masked_data(var_name, rm_assessments=assessment)
        assert np.sum(np.where(qc_ma.mask)[0]) == index_sum

    qc_ma = ds.qcfilter.get_masked_data(var_name, rm_assessments=['Bucky'])
    assert np.sum(np.where(qc_ma.mask)[0]) == 0

    qc_ma = ds.qcfilter.get_masked_data(var_name, rm_assessments=['Boomer', 'Sooner'])
    assert np.sum(np.where(qc_ma.mask)[0]) == 94850

    qc_ma = ds.qcfilter.get_masked_data(
        var_name,
        rm_assessments=['Boomer', 'Sooner', 'Indeterminate', 'Suspect', 'Bad', 'Incorrect'],
    )
    assert np.sum(np.where(qc_ma.mask)[0]) == 244650

    del ds


def test_qc_summary_scalar():
    # Test scalar variables. Currently not implemented so just check that we
    # don't do anything.
    var_names = ['alt', 'temp_mean']
    ds = read_arm_netcdf(EXAMPLE_MET1, keep_variables=var_names)

    test_meanings = ['Testing Incorrect', 'Testing Suspect']
    test_assessments = ['Incorrect', 'Suspect']

    for var_name in var_names:
        for ii, _ in enumerate(test_assessments):
            ds.qcfilter.add_test(
                var_name,
                index=0,
                test_meaning=test_meanings[ii],
                test_assessment=test_assessments[ii],
            )

    with pytest.warns(UserWarning, match="Unable to process scalar variable"):
        ds = ds.qcfilter.create_qc_summary(normalize_assessment=False)

    assert 'flag_masks' in ds[f'qc_{var_names[0]}'].attrs.keys()
    assert 'flag_values' not in ds[f'qc_{var_names[0]}'].attrs.keys()
    assert 'flag_masks' not in ds[f'qc_{var_names[1]}'].attrs.keys()
    assert 'flag_values' in ds[f'qc_{var_names[1]}'].attrs.keys()


@pytest.mark.big
@pytest.mark.skipif('ARCHIVE_DATA' not in environ, reason="Running outside ADC system.")
def test_qc_summary_big_data():
    """
    We want to test on as much ARM data as possible. But we do not want to force
    a large amount of test data in GitHub. Plan is to see if the pytest code is being
    run on ARM system and if so then run on historical data. If running on GitHub
    then don't run tests. Also, have a switch to not force this big test to always
    run as that would be mean to the developer. So need to periodicaly run with the
    manual switch enabled.

    All exceptions are caught and a file name is sent to the output file when
    an exception is found. Since this is testing 10,000+ files it will take hours
    to run. I suggest you run in background and capture the standard out to a different
    file. If no files are written to the output file then all tests passed.

    Output file name follows the convention of:
        ~/test_qc_summary_big_data.{datetime}.txt

    To Run this test set keyword on pytest command line:
    > pytest -s --runbig test_qc_summary.py::test_qc_summary_big_data &> ~/out.txt &


    """

    base_path = Path(environ['ARCHIVE_DATA'])
    if not base_path.is_dir():
        return

    # Set number of files from each directory to test.
    skip_sites = [
        'shb',
        'wbu',
        'dna',
        'rld',
        'smt',
        'nic',
        'isp',
        'dmf',
        'nac',
        'rev',
        'yeu',
        'zrh',
        'osc',
    ]
    skip_datastream_codes = ['mmcrmom']
    num_files = 3
    expected_assessments = ['Not failing', 'Suspect', 'Indeterminate', 'Incorrect', 'Bad']

    testing_files = []

    if len(testing_files) == 0:
        filename = (
            f'test_qc_summary_big_data.{datetime.datetime.utcnow().strftime("%Y%m%d.%H%M%S")}.txt'
        )
        output_file = Path(environ['HOME'], filename)
        output_file.unlink(missing_ok=True)
        output_file.touch()

        site_dirs = list(base_path.glob('???'))
        for site_dir in site_dirs:
            if site_dir.name in skip_sites:
                continue

            datastream_dirs = list(site_dir.glob('*.[bc]?'))
            for datastream_dir in datastream_dirs:
                if '-' in datastream_dir.name:
                    continue

                fn_obj = DatastreamParserARM(datastream_dir.name)
                facility = fn_obj.facility
                if facility is not None and facility[0] in ['A', 'X', 'U', 'F', 'N']:
                    continue

                datastream_class = fn_obj.datastream_class
                if datastream_class is not None and datastream_class in skip_datastream_codes:
                    continue

                files = list(datastream_dir.glob('*.nc'))
                files.extend(datastream_dir.glob('*.cdf'))
                if len(files) == 0:
                    continue

                num_tests = num_files
                if len(files) < num_files:
                    num_tests = len(files)

                for ii in range(0, num_tests):
                    testing_files.append(random.choice(files))

        print(f"\nTesting {len(testing_files)} files\n")
        print(f"Output file name = {output_file}\n")

    for file in testing_files:
        try:
            print(f"Testing: {file}")
            ds = read_arm_netcdf(str(file), cleanup_qc=True, decode_times=False)
            ds = ds.qcfilter.create_qc_summary()

            created_qc_summary = False
            for var_name in ds.data_vars:
                qc_var_name = ds.qcfilter.check_for_ancillary_qc(
                    var_name, add_if_missing=False, cleanup=False
                )

                if qc_var_name is None:
                    continue

                created_qc_summary = True
                assert isinstance(ds[qc_var_name].attrs['flag_values'], list)
                assert isinstance(ds[qc_var_name].attrs['flag_assessments'], list)
                assert isinstance(ds[qc_var_name].attrs['flag_meanings'], list)
                assert len(ds[qc_var_name].attrs['flag_values']) >= 1
                assert len(ds[qc_var_name].attrs['flag_assessments']) >= 1
                assert len(ds[qc_var_name].attrs['flag_meanings']) >= 1
                assert ds[qc_var_name].attrs['flag_assessments'][0] == 'Not failing'
                assert (
                    ds[qc_var_name].attrs['flag_meanings'][0] == 'Not failing quality control tests'
                )

                for assessment in ds[qc_var_name].attrs['flag_assessments']:
                    assert assessment in expected_assessments

            if created_qc_summary:
                assert "Quality control summary implemented by ACT" in ds.attrs['history']

            del ds

        except Exception:
            with open(output_file, "a") as myfile:
                myfile.write(f"{file}\n")



"""
Xarray Plotting Examples
------------------------

This is an example of how to use some different aspects
of ACT's plotting tools as well as Xarray's tools.

"""

import matplotlib.pyplot as plt
from arm_test_data import DATASETS

import act

# Set up plot space ahead of time
fig, ax = plt.subplots(3, figsize=(10, 7))

# Plotting up high-temporal resolution 2D data can be very slow at times.
# In order to increase the speed, the data can be resampled to a courser
# resolution prior to plotting.  Using Xarray's resample and selecting
# the nearest neighbor will greatly increase the speed.
filename_ceil = DATASETS.fetch('sgpceilC1.b1.20190101.000000.nc')
ds = act.io.arm.read_arm_netcdf(filename_ceil)
ds = ds.resample(time='1min').nearest()

# These data can be plotted up using the existing xarray functionality
# which is quick and easy
ds['backscatter'].plot(x='time', ax=ax[0])

# or using ACT
display = act.plotting.TimeSeriesDisplay(ds)
display.assign_to_figure_axis(fig, ax[1])
display.plot('backscatter')

# When using ACT, the axis object can also be manipulated using normal
# matplotlib calls for more personalized customizations
display = act.plotting.TimeSeriesDisplay(ds)
display.assign_to_figure_axis(fig, ax[2])
display.plot('backscatter')
display.axes[-1].set_ylim([0, 1500])

plt.show()



"""
Using ACT for Satellite data
----------------------------

Simple example for working with satellite data.
It is recommended that users explore other
satellite specific libraries suchas SatPy.

Author: Adam Theisen
"""

import matplotlib.pyplot as plt
import numpy as np
from arm_test_data import DATASETS

import act

# Read in VISST Data
files = DATASETS.fetch('enavisstgridm11minnisX1.c1.20230307.000000.cdf')
ds = act.io.read_arm_netcdf(files)

# Plot up the VISST cloud percentage using XSectionDisplay
display = act.plotting.XSectionDisplay(ds, figsize=(10, 8))
display.plot_xsection_map(
    'cloud_percentage', x='longitude', y='latitude', isel_kwargs={'time': 0, 'cld_type': 0}
)
plt.show()

# Download ARM TSI Data
files = DATASETS.fetch('enatsiskycoverC1.b1.20230307.082100.cdf')
ds_tsi = act.io.read_arm_netcdf(files)
ds_tsi = ds_tsi.where(ds_tsi.percent_opaque > 0)
ds_tsi = ds_tsi.resample(time='30min').mean()

# Set coordinates to extra data for ENA
ena_lat = 39.091600
ena_lon = 28.025700

lat = ds['lat'].values
lon = ds['lon'].values

# Find the nearest pixel for the satellite data and extract it
lat_ind = np.argmin(np.abs(lat - ena_lat))
lon_ind = np.argmin(np.abs(lon - ena_lon))

ds_new = ds.isel(lat=lat_ind, lon=lon_ind, cld_type=0)

# Plot the comparison using TimeSeriesDisplay
display = act.plotting.TimeSeriesDisplay({'Satellite': ds_new, 'ARM': ds_tsi}, figsize=(15, 8))
display.plot('cloud_percentage', dsname='Satellite', label='VISST Cloud Percentage')
display.plot('percent_opaque', dsname='ARM', label='ARM TSI Percent Opaque')
display.day_night_background(dsname='ARM')
plt.legend()
plt.show()



"""
Transformations and QC
----------------------

Built-in transformations using xarray are not
quality-control aware. This example shows how
a user should apply QC prior to performing transformations.

"""

import matplotlib.pyplot as plt
from arm_test_data import DATASETS

import act

# Read in some sample MFRSR data and clean up the QC
filename_mfrsr = DATASETS.fetch('sgpmfrsr7nchE11.b1.20210329.070000.nc')
ds = act.io.arm.read_arm_netcdf(filename_mfrsr, cleanup_qc=True)

# Let's resample the data to 5 minutes and take the mean
ds_5min = ds.resample(time='5min').mean()

variable = 'diffuse_hemisp_narrowband_filter4'

# Let's look at a before and after of one of the qc variables
print('With no QC applied before transformation')
print('Before (10 1-minute samples): ', ds['qc_' + variable].values[0:10])
print('After: (2 5-minute averages)', ds_5min['qc_' + variable].values[0:2])

# That new QC variable does not make sense at all and should be an int
# What needs to happen is that we apply QC as the user see's fit to all
# variables before the transformations take place.
print('\nAverage of ', variable, ' before and after applying QC')
print('Note the change in the second value')
print('Before (2 5 - minute averages): ', ds[variable].values[0:2])

ds.qcfilter.datafilter(rm_assessments=['Bad', 'Indeterminate'])
ds_5minb = ds.resample(time='5min').mean()

# Print out the corresponding variable values
print('After: (2 5 - minute averages)', ds_5minb[variable].values[0:2])

## Plot up the variable and qc block plot
display = act.plotting.TimeSeriesDisplay(
    {'Original': ds, 'Average': ds_5min, 'Average_QCd': ds_5minb},
    figsize=(15, 10),
    subplot_shape=(2,),
)
display.plot(variable, dsname='Original', subplot_index=(0,), day_night_background=True)
display.plot(
    variable, dsname='Average', subplot_index=(1,), day_night_background=True, label='No QC'
)
display.plot(
    variable, dsname='Average_QCd', subplot_index=(1,), day_night_background=True, label='QC'
)
plt.legend()
plt.show()



import numpy as np
from arm_test_data import DATASETS

import act


def test_calculate_gradient_pbl():
    # Read and apply connections
    ds = act.io.arm.read_arm_netcdf(DATASETS.fetch('sgpceilC1.b1.20190101.000000.nc'))
    ds = act.corrections.correct_ceil(ds, var_name='backscatter')

    # Call the Retrieval
    ds = act.retrievals.pbl_lidar.calculate_gradient_pbl(
        ds, parm="backscatter", smooth_dis=3, min_height=200
    )

    # create a subset for testing
    subset = ds.sel(time=slice("2019-01-01T11:30:00", "2019-01-01T11:40:00"))
    # Test the mean of the profile for the subset time
    np.testing.assert_array_almost_equal(subset.pbl_gradient.mean(), 436.875, decimal=3)
    # Test the minimum PBL Height during the period
    #   note this will test the minimum height threshold assigned
    np.testing.assert_almost_equal(subset.pbl_gradient.min(), 225.0, 1)

    # test attributes
    assert ds['pbl_gradient'].attrs["input_parameter"] == "backscatter"
    assert ds['pbl_gradient'].attrs["units"] == "m"


def test_calculate_modified_gradient_pbl():
    # Read and apply connections
    ds = act.io.arm.read_arm_netcdf(DATASETS.fetch('sgpceilC1.b1.20190101.000000.nc'))
    ds = act.corrections.correct_ceil(ds, var_name='backscatter')

    # Call the Retrieval
    ds = act.retrievals.pbl_lidar.calculate_modified_gradient_pbl(
        ds, parm="backscatter", smooth_dis=3, min_height=200, threshold=1e-5
    )

    # create a subset for testing
    subset = ds.sel(time=slice("2019-01-01T11:30:00", "2019-01-01T11:40:00"))
    # Test the mean of the profile for the subset time
    np.testing.assert_array_almost_equal(subset.pbl_mod_gradient.mean(), 372.631, decimal=3)
    # Test the minimum PBL Height during the period
    #   note this will test the minimum height threshold assigned
    np.testing.assert_almost_equal(subset.pbl_mod_gradient.min(), 225.0, 1)

    # test attributes
    assert ds['pbl_mod_gradient'].attrs["input_parameter"] == "backscatter"
    assert ds['pbl_mod_gradient'].attrs["units"] == "m"



"""
This module contains procedures for exploring and downloading data
from a variety of web services

"""

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submodules=[
        'ameriflux',
        'arm',
        'asos',
        'airnow',
        'cropscape',
        'improve',
        'noaapsl',
        'neon',
        'surfrad',
        'nasa',
    ],
    submod_attrs={
        'ameriflux': ['download_ameriflux_data'],
        'arm': ['download_arm_data', 'download_arm_data_mod', 'get_arm_doi'],
        'asos': ['get_asos_data'],
        'airnow': ['get_airnow_bounded_obs', 'get_airnow_obs', 'get_airnow_forecast'],
        'cropscape': ['get_crop_type'],
        'improve': ['get_improve_data'],
        'noaapsl': ['download_noaa_psl_data'],
        'neon': ['get_neon_site_products', 'get_neon_product_avail', 'download_neon_data'],
        'surfrad': ['download_surfrad_data'],
        'nasa': ["download_mplnet_data", "get_mplnet_meta"],
    },
)



import pytest

import act


def test_decode_present_weather():
    ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_MET1)
    ds = act.utils.decode_present_weather(ds, variable='pwd_pw_code_inst')

    data = ds['pwd_pw_code_inst_decoded'].values
    result = 'No significant weather observed'
    assert data[0] == result
    assert data[100] == result
    assert data[600] == result

    pytest.raises(ValueError, act.utils.inst_utils.decode_present_weather, ds)
    pytest.raises(
        ValueError,
        act.utils.inst_utils.decode_present_weather,
        ds,
        variable='temp_temp',
    )



"""
This module contains corrections for micropulse lidars

"""

import warnings

import numpy as np
import xarray as xr


def correct_mpl(
    ds,
    co_pol_var_name='signal_return_co_pol',
    cross_pol_var_name='signal_return_cross_pol',
    co_pol_afterpuls_var_name='afterpulse_correction_co_pol',
    cross_pol_afterpulse_var_name='afterpulse_correction_cross_pol',
    overlap_corr_var_name='overlap_correction',
    overlap_corr_heights_var_name='overlap_correction_heights',
    range_bins_var_name='range_bins',
    height_var_name='height',
    ratio_var_name='cross_co_ratio',
):
    """
    This procedure corrects MPL data:
    1.) Throw out data before laser firing (heights < 0).
    2.) Remove background signal.
    3.) Afterpulse Correction - Subtraction of (afterpulse-darkcount).
    NOTE: Currently the Darkcount in VAPS is being calculated as
    the afterpulse at ~30km. But that might not be absolutely
    correct and we will likely start providing darkcount profiles
    ourselves along with other corrections.
    4.) Range Correction.
    5.) Overlap Correction (Multiply).

    If the height variable changes between netCDF files, Xarray will turn the height
    dimention variables from a 1D to a 2D array. This will cause issues with processing
    and other data manipulation. To fix this the 2D height will be converted to a 1D
    array by using the median value for each height value.

    Note: Deadtime and darkcount corrections are not being applied yet.

    Parameters
    ----------
    ds : xarray.Dataset
        The Xarray Dataset containing data
    co_pol_var_name : str
        The Co-polar variable name used in Dataset
    cross_pol_var_name : str
        The Cross-polar variable name used in Dataset
    co_pol_afterpuls_var_name : str
        The Co-polar after pulse variable name used in Dataset
    cross_pol_afterpulse_var_name : str
        The Cross-polar afterpulse variable name used in Dataset
    overlap_corr_var_name : str
        The overlap correction variable name used in Dataset
    overlap_corr_heights_var_name : str
        The overlap correction height variable name used in Dataset
    range_bins_var_name : str
        The range bins variable name used in Dataset
    height_var_name : str
        The height variable name used in Dataset
    ratio_var_name : str
        The variable name to use for newly created variable of co/cross ratio.
        Set to None if do not want this new variables created in Dataset.

    Returns
    -------
    ds : xarray.Dataset
        Xarray dataset containing the corrected values. The original Xarray Dataset
        passed in is modified.

    """

    data_dims = ds[co_pol_var_name].dims

    # Overlap Correction Variable
    op = ds[overlap_corr_var_name].values
    if len(op.shape) > 1:
        op = op[0, :]

    op_height = ds[overlap_corr_heights_var_name].values
    if len(op_height.shape) > 1:
        op_height = op_height[0, :]

    # Check if height has dimentionality of time and height. If so reduce
    # height to only dimentionality of height in the dataset before removing
    # values less than 0.
    if len(ds[height_var_name].shape) > 1:
        reduce_dim_name = {'time'} & set(ds[height_var_name].dims)
        ds[height_var_name] = ds[height_var_name].reduce(
            func=np.median, dim=reduce_dim_name, keep_attrs=True
        )

    # 1 - Remove negative height data
    ds = ds.where(ds[height_var_name].load() > 0.0, drop=True)
    height = ds[height_var_name].values

    # Get indices for calculating background
    if len(ds.height.shape) > 1:
        ind = [ds.height.shape[1] - 50, ds.height.shape[1] - 2]
    else:
        ind = [ds.height.shape[0] - 50, ds.height.shape[0] - 2]

    # Subset last gates into new dataset
    dummy = ds.isel(range_bins=xr.DataArray(np.arange(ind[0], ind[1])))

    # Turn off warnings
    warnings.filterwarnings('ignore')

    # Run through co and cross pol data for corrections
    co_bg = dummy[co_pol_var_name]
    co_bg = co_bg.where(co_bg.load() > -9998.0)
    co_bg = co_bg.mean(dim='dim_0').values

    x_bg = dummy[cross_pol_var_name]
    x_bg = x_bg.where(x_bg.load() > -9998.0)
    x_bg = x_bg.mean(dim='dim_0').values

    # Seems to be the fastest way of removing background signal at the moment
    co_data = ds[co_pol_var_name].where(ds[co_pol_var_name].load() > 0).values
    x_data = ds[cross_pol_var_name].where(ds[cross_pol_var_name].load() > 0).values
    for i in range(len(ds['time'].values)):
        co_data[i, :] = co_data[i, :] - co_bg[i]
        x_data[i, :] = x_data[i, :] - x_bg[i]

    # After Pulse Correction Variable
    co_ap = ds[co_pol_afterpuls_var_name].values
    # Fix dimentionality if backwards
    co_ap_dims = ds[co_pol_afterpuls_var_name].dims
    if len(co_ap_dims) > 1 and co_ap_dims[::-1] == data_dims:
        co_ap = np.transpose(co_ap)

    x_ap = ds[cross_pol_afterpulse_var_name].values
    # Fix dimentionality if backwards
    x_ap_dims = ds[cross_pol_afterpulse_var_name].dims
    if len(x_ap_dims) > 1 and x_ap_dims[::-1] == data_dims:
        x_ap = np.transpose(x_ap)

    # Afterpulse Correction
    co_data = co_data - co_ap
    x_data = x_data - x_ap

    # R-Squared Correction
    co_data = co_data * height**2
    x_data = x_data * height**2

    # Overlap Correction
    for j in range(ds[range_bins_var_name].size):
        if len(height.shape) > 1:
            idx = (np.abs(op_height - height[0, j])).argmin()
        else:
            # Overlap Correction
            idx = (np.abs(op_height - height[j])).argmin()

        co_data[:, j] = co_data[:, j] * op[idx]
        x_data[:, j] = x_data[:, j] * op[idx]

    # Create the co/cross ratio variable
    if ratio_var_name is not None:
        ratio = (x_data / (x_data + co_data)) * 100.0
        ds[ratio_var_name] = ds[co_pol_var_name].copy(data=ratio)
        ds[ratio_var_name].attrs['long_name'] = 'Cross-pol / Co-pol ratio * 100'
        ds[ratio_var_name].attrs['units'] = '1'
        try:
            del ds[ratio_var_name].attrs['ancillary_variables']
            del ds[ratio_var_name].attrs['description']
        except KeyError:
            pass

    # Convert data to decibels
    co_data = 10.0 * np.log10(co_data)
    x_data = 10.0 * np.log10(x_data)

    # Write data to Xarray dataset
    ds[co_pol_var_name].values = co_data
    ds[cross_pol_var_name].values = x_data

    # Update units
    ds[co_pol_var_name].attrs['units'] = f"10 * log10({ds[co_pol_var_name].attrs['units']})"
    ds[cross_pol_var_name].attrs['units'] = f"10 * log10({ds[cross_pol_var_name].attrs['units']})"

    return ds



import gzip
import shutil
import tarfile
import tempfile
import types
from os import PathLike, sep
from pathlib import Path
from shutil import rmtree

import numpy as np
import pandas as pd

import act

try:
    import moviepy.video.io.ImageSequenceClip
    from moviepy import VideoFileClip

    MOVIEPY_AVAILABLE = True
except (ImportError, RuntimeError):
    MOVIEPY_AVAILABLE = False


def pack_tar(filenames, write_filename=None, write_directory=None, remove=False):
    """
    Creates TAR file from list of filenames provided. Currently only works with
    all files existing in the same directory.

    Parameters
    ----------
    filenames : str or list
        Filenames to be placed in TAR file
    write_filename : str, pathlib.Path, None
        TAR output filename. If not provided will use file name 'created_tarfile.tar'
    write_directory : str, pathlib.Path, None
        Path to directory to write TAR file. If the directory does not exist will
        be created.
    remove : boolean
        Delete provided filenames after making TAR file

    Returns
    -------
    list
        List of files extracted from the TAR file or full path to created direcotry
        containing extracted files.

    """

    if write_filename is None:
        write_filename = 'created_tarfile.tar'

    if isinstance(filenames, (str, PathLike)):
        filenames = [filenames]

    if write_directory is not None:
        write_directory = Path(write_directory)
        write_directory.mkdir(parents=True, exist_ok=True)
        write_filename = Path(write_filename).name
    elif Path(write_filename).parent != Path('.'):
        write_directory = Path(write_filename).parent
    else:
        write_directory = Path('.')

    if not str(write_filename).endswith('.tar'):
        write_filename = str(write_filename) + '.tar'

    write_filename = Path(write_directory, write_filename)
    tar_file_handle = tarfile.open(write_filename, "w")
    for filename in filenames:
        tar_file_handle.add(filename, arcname=Path(filename).name)

    tar_file_handle.close()

    if remove:
        for filename in filenames:
            Path(filename).unlink()

    return str(write_filename)


def unpack_tar(
    tar_files, write_directory=None, temp_dir=False, randomize=True, return_files=True, remove=False
):
    """
    Unpacks TAR file contents into provided base directory

    Parameters
    ----------
    tar_files : str or list
        path to TAR file to be unpacked
    write_directory : str or pathlib.Path
        base path to extract contents of TAR files or create a new randomized directory
        to extract contents of TAR file.
    temp_dir : boolean
        Should a temporary directory be created and TAR files extracted to the new directory.
        write_directory and randomize are ignored if this option is used.
    randomize : boolean
        Create a new randomized directory to extract TAR files into.
    return_files : boolean
        When set will return a list of full path filenames to the extracted files.
        When set to False will return full path to directory containing extracted files.
    remove : boolean
        Delete provided TAR files after extracting files.

    Returns
    -------
    files : list or str
        List of full path files extracted from the TAR file or full path to direcotry
        containing extracted files.

    """

    files = []

    if isinstance(tar_files, (str, PathLike)):
        tar_files = [tar_files]

    out_dir = Path.cwd()
    if temp_dir is True:
        out_dir = Path(tempfile.TemporaryDirectory().name)
    else:
        if write_directory is not None:
            out_dir = Path(write_directory)
        else:
            out_dir = Path(Path(tar_files[0]).parent)

        if out_dir.is_dir() is False:
            out_dir.mkdir(parents=True, exist_ok=True)

        if randomize:
            out_dir = Path(tempfile.mkdtemp(dir=out_dir))

    for tar_file in tar_files:
        try:
            tar = tarfile.open(tar_file)
            tar.extractall(path=out_dir, filter='data')
            result = [str(Path(out_dir, ii.name)) for ii in tar.getmembers()]
            files.extend(result)
            tar.close()
        except tarfile.ReadError:
            print("Could not extract files from the tar_file")

    if return_files is False:
        files = str(out_dir)
    else:
        files.sort()

    if remove:
        for tar_file in tar_files:
            Path(tar_file).unlink()

    return files


def cleanup_files(dirname=None, files=None):
    """
    Cleans up files and directory possibly created from unpacking TAR files with unpack_tar()

    Parameters
    ----------
    dirname : str, pathlib.Path, None
        Path to directory of extracted files which will be removed.
    files : str, pahtlib.Path, list, None
        Full path file name(s) from extracted TAR file.
        Assumes the directory this file exists in should be removed.

    """

    if isinstance(files, (str, PathLike)):
        files = [str(files)]

    try:
        if dirname is not None:
            rmtree(dirname)

        if files is not None and len(files) > 0 and Path(files[0]).is_file():
            out_dir = Path(files[0]).parent
            rmtree(out_dir)

    except Exception as error:
        print("\nError removing files:", error)


def is_gunzip_file(filepath):
    """
    Function to test if file is a gunzip file.

    Parameters
    ----------

    filepath : str or pathlib.Path to file to test

    Returns
    -------
    test : boolean
        Result from testing if file is a gunzip file

    """

    try:
        with open(str(filepath), 'rb') as test_f:
            return test_f.read(2) == b'\x1f\x8b'
    except Exception:
        return False


def pack_gzip(filename, write_directory=None, remove=False):
    """
    Creates a gunzip file from a filename path

    Parameters
    ----------
    filename : str, pathlib.Path
        Filename to use in creation of gunzip version.
    write_directory : str, pahtlib.Path, list, None
        Path to directory to place newly created gunzip file.
    remove : boolean
        Remove provided filename after creating gunzip file

    Returns
    -------
    write_filename : str
        Full path name of created gunzip file

    """

    write_filename = Path(filename).name + '.gz'

    if write_directory is not None:
        write_filename = Path(write_directory, write_filename)
        Path(write_directory).mkdir(parents=True, exist_ok=True)
    else:
        write_filename = Path(Path(filename).parent, write_filename)

    with open(filename, 'rb') as f_in:
        with gzip.open(write_filename, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)

    if remove:
        Path(filename).unlink()

    return str(write_filename)


def unpack_gzip(filename, write_directory=None, remove=False):
    """
    Extracts file from a gunzip file.

    Parameters
    ----------
    filename : str, pathlib.Path
        Filename to use in extraction of gunzip file.
    write_directory : str, pahtlib.Path, list, None
        Path to directory to place newly created gunzip file.
    remove : boolean
        Remove provided filename after creating gunzip file

    Returns
    -------
    write_filename : str
        Full path name of created gunzip file

    """

    if write_directory is None:
        write_directory = Path(filename).parent

    write_filename = Path(filename).name
    if write_filename.endswith('.gz'):
        write_filename = write_filename.replace(".gz", "")

    write_filename = Path(write_directory, write_filename)

    with gzip.open(filename, "rb") as f_in:
        with open(write_filename, "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)

    if remove:
        Path(filename).unlink()

    return str(write_filename)


def generate_movie(images, write_filename=None, fps=10, duration=None, **kwargs):
    """
    Creates a movie from a list of images or convert movie to different type

    Parameters
    ----------
    images : list, PosixPath generator, path to a directory, single string/PosixPath to movie
        List of images in the correct order to make into a movie or a generator from
        a pathlib.Path.glob() search. If a path to directory will create movie from all files
        in that directory in alpanumeric order. If a single file to a movie will allow for converting
        to new format defined by file extension of write_filename.
    write_filename : str, pathlib.Path, None
        Movie output filename. Default is 'movie.mp4' in current directory. If a path to a directory
        that does not exist, will create the directory path.
    fps: int
        Frames per second. Passed into moviepy->ImageSequenceClip() method
    duration : float, int or None
        Converting mpg format can have issues with reading the duration of the movie. Set
        to number of seconds to override the derived value if the result is not what you expect.
    **kwargs: dict
        Optional keywords passed into moviepy->write_videofile() method


    Returns
    -------
    write_filename : str
        Full path name of created movie file

    """
    if not MOVIEPY_AVAILABLE:
        raise ImportError('MoviePy v2.X needs to be installed on your system to make movies.')

    if int(moviepy.__version__.split('.')[0]) <= 1:
        raise ImportError('MoviePy v2.X needs to be installed on your system to make movies.')

    # Set default movie name
    if write_filename is None:
        write_filename = Path(Path().cwd(), 'movie.mp4')

    # Check if images is pointing to a directory. If so ensure is a string not PosixPath
    IS_MOVIE = False
    if isinstance(images, (types.GeneratorType, list, tuple)):
        images = [str(image) for image in images]
        images.sort()
    elif isinstance(images, (PathLike, str)) and Path(images).is_file():
        IS_MOVIE = True
        images = str(images)
    elif isinstance(images, (PathLike, str)) and Path(images).is_dir():
        images = str(images)

    # Ensure full path to filename exists
    write_directory = Path(write_filename).parent
    write_directory.mkdir(parents=True, exist_ok=True)

    if IS_MOVIE:
        with VideoFileClip(images) as clip:
            # There can be an issue converting mpeg to other movie format because the
            # duration parameter in the movie file is not set. So moviepy guesses and
            # can get the duration wrong.
            if duration is not None:
                clip = clip.with_start(0)
                clip = clip.with_duration(duration)
                clip = clip.with_end(duration)

            clip.write_videofile(str(write_filename), **kwargs)

    else:
        clip = moviepy.video.io.ImageSequenceClip.ImageSequenceClip(images, fps=fps)
        clip.write_videofile(str(write_filename), **kwargs)

    return str(write_filename)


def arm_standards_validator(file=None, dataset=None, verbose=True):
    """
    ARM Data Validator (ADV) - Checks to ensure that ARM standards are being followed
    in the files or dataset passed to it.  Note, this includes a minimal set of
    standards that it checks against

    Parameters
    ----------
    file : str
        Filename to check against ARM standards.  Do not pass in both a file and dataset
    dataset : xarray.DataSet
        Xarray dataset of an already read in file.
    verbose : boolean
        Defaults to print out errors in addition to returning a list of them

    Returns
    -------
    err : list
        List of errors in the data

    """

    # Set up the error tracking list
    err = []
    if file is not None and isinstance(file, str):
        # Check file naming standards
        if len(file.split(sep)[-1]) > 60.0:
            err.append('Filename length exceeds 60 characters')
        try:
            f_obj = act.utils.data_utils.DatastreamParserARM(file)
        except Exception as e:
            print(e)

        if (
            (f_obj.site is None)
            or (f_obj.datastream_class is None)
            or (f_obj.level is None)
            or (f_obj.facility is None)
            or (f_obj.date is None)
            or (f_obj.time is None)
            or (f_obj.ext is None)
        ):
            err.append(
                'Filename does not follow the normal ARM convention: '
                + '(sss)(inst)(qualifier)(temporal)(Fn).(dl).(yyyymmdd).(hhmmss).nc'
            )
        else:
            if f_obj.level[0] not in ['0', 'a', 'b', 'c', 's', 'm']:
                err.append(f_obj.level + ' is not a standard ARM data level')

        results = act.utils.arm_site_location_search(
            site_code=f_obj.site, facility_code=f_obj.facility
        )
        if len(results) == 0:
            err.append('Site and facility are not ARM standard')

    # The ability to read a file from NetCDF to xarray will catch a lot of the
    # problems with formatting.  This would leave standard ARM checks
    try:
        if dataset is None and file is not None:
            ds = act.io.read_arm_netcdf(file)
        elif dataset is not None:
            ds = dataset
        else:
            raise ValueError('File and dataset are both None')
    except Exception as e:
        return ['File is not in a standard format that is readable by xarray: ' + str(e)]

    # Review time variables for errors for conformance to standards
    if 'time' not in list(ds.dims)[0]:
        err.append('"time" is required to be the first dimension.')

    for c in list(ds.coords):
        if c not in ds.dims:
            err.append(c + ': Coordinate is not included in dimensions.')

    if any(np.isnan(ds['time'].values)):
        err.append('Time must not include NaNs.')

    duplicates = sum(ds['time'].to_pandas().duplicated())
    if duplicates > 0:
        err.append('Duplicate times present in the file')

    diff = ds['time'].diff('time')
    idx = np.where(diff <= pd.Timedelta(0))
    if len(idx[0]) > 0:
        err.append('Time is not in increasing order')

    if 'base_time' not in ds or 'time_offset' not in ds:
        err.append('ARM requires base_time and time_offset variables.')

    # Check to make sure other coordinate variables don't have nans
    # Also check to make sure coordinate variables are not decreasing
    if len(list(ds.coords)) > 1:
        for d in ds.coords:
            if d == 'time':
                continue
            if any(np.isnan(ds[d].values)):
                err.append('Coordinates must not include NaNs ' + d)

            diff = ds[d].diff(d)
            idx = np.where(diff <= 0.0)
            if len(idx[0]) > 0:
                err.append(d + ' is not in increasing order')
            if 'missing_value' in ds[d].encoding:
                err.append(d + ' should not include missing value')

    # Verify that each variable has a long_name and units attributes
    for v in ds:
        if (len(ds[v].dims) > 0) and ('time' not in list(ds[v].dims)[0]) and ('bounds' not in v):
            err.append(v + ': "time" is required to be the first dimension.')
        if (ds[v].size == 1) and (len(ds[v].dims) > 0):
            err.append(v + ': is not defined as a scalar.')
        if 'long_name' not in ds[v].attrs:
            err.append('Required attribute long_name not in ' + v)
        else:
            if not ds[v].attrs['long_name'][0].isupper():
                err.append(v + ' long_name attribute does not start with uppercase')

        if (
            ('qc_' not in v)
            and (v not in ['time', 'time_offset', 'base_time', 'lat', 'lon', 'alt'])
            and ('bounds' not in v)
        ):
            if ('missing_value' not in ds[v].encoding) and ('FillValue' not in ds[v].encoding):
                err.append(v + ' does not include missing_value or FillValue attributes')

        # QC variable checks
        if 'qc_' in v:
            if v[3:] not in ds:
                err.append('QC variable does not have a corresponding variable ' + v[3:])
            if 'ancillary_variables' not in ds[v[3:]].attrs:
                err.append(
                    v[3:] + ' does not include ancillary_variable attribute pointing to ' + v
                )
            if 'description' not in ds[v].attrs:
                err.append(v + ' does not include description attribute')
            if 'flag_method' not in ds[v].attrs:
                err.append(v + ' does not include flag_method attribute')

        if (v not in ['base_time', 'time_offset']) and ('bounds' not in v):
            if 'units' not in ds[v].attrs:
                err.append('Required attribute units not in ' + v)

    # Lat/Lon/Alt Checks
    if 'lat' not in ds:
        err.append('ARM requires the latitude variable to be named lat')
    else:
        if 'standard_name' in ds['lat'].attrs:
            if ds['lat'].attrs['standard_name'] != 'latitude':
                err.append('ARM requires the lat standard_name to be latitude')
        else:
            err.append('"lat" variable does not have a standard_name attribute')
    if 'lon' not in ds:
        err.append('ARM requires the longitude variable to be named lon')
    else:
        if 'standard_name' in ds['lon'].attrs:
            if ds['lon'].attrs['standard_name'] != 'longitude':
                err.append('ARM requires the lon standard_name to be longitude')
        else:
            err.append('"long" variable does not have a standard_name attribute')
    if 'alt' not in ds:
        err.append('ARM requires the altitude variable to be named alt')
    else:
        if 'standard_name' in ds['alt'].attrs:
            if ds['alt'].attrs['standard_name'] != 'altitude':
                err.append('ARM requires the alt standard_name to be altitude')
        else:
            err.append('"alt" variable does not have a standard_name attribute')

    # Required global attributes
    req_att = ['doi', 'sampling_interval', 'averaging_interval']
    for ra in req_att:
        if ra not in ds.attrs:
            err.append('Global attribute is missing: ' + ra)

    if verbose:
        if len(err) > 0:
            [print(e) for e in err]
        else:
            print('File is passing standards checks')

    return err



import act


def test_generic_sobel_cbh():
    ceil = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_CEIL1)

    ceil = ceil.resample(time='1min').nearest()
    ceil = act.retrievals.cbh.generic_sobel_cbh(
        ceil,
        variable='backscatter',
        height_dim='range',
        var_thresh=1000.0,
        fill_na=0.0,
        edge_thresh=5,
    )
    cbh = ceil['cbh_sobel_backscatter'].values
    assert cbh[500] == 615.0
    assert cbh[1000] == 555.0
    ceil.close()



"""
Query the ARM DQR webservice
----------------------------

This example shows how to query the ARM Data Quality Report (DQR) webservice,
to retrieve the machine readable DQR information.  This function adds the
DQR information to corresponding QC variables (qc_[variablename]) in the
dataset.  The example then shows how to visualize this information as we
do with the normal QC variables.

Author: Adam Theisen
"""

from arm_test_data import DATASETS
from matplotlib import pyplot as plt

import act

# Read in sample AOSMET data
filename_aosmet = DATASETS.fetch('maraosmetM1.a1.20180201.000000.nc')
ds = act.io.arm.read_arm_netcdf(filename_aosmet)

# Query DQR webservice for a specific variable
# As can be seen in the "Plotting QC Flags" example
# a call to obj.clean.cleanup() would normally be needed
# in order to plot up ARM's QC information.  In this case
# the call to add DQRs to the QC automatically applies that
# cleanup so you don't have to.
variable = 'temperature_ambient'
ds = act.qc.arm.add_dqr_to_qc(ds, variable=variable)

# Plot data
# Creat Plot Display
display = act.plotting.TimeSeriesDisplay(ds, figsize=(15, 10), subplot_shape=(2,))

# Plot temperature data in top plot
display.plot(variable, subplot_index=(0,))

# Plot QC data
display.qc_flag_block_plot(variable, subplot_index=(1,))
plt.show()



import os

import act


def test_neon():
    site_code = 'BARR'
    result = act.discovery.get_neon_site_products(site_code, print_to_screen=True)
    assert 'DP1.00002.001' in result
    assert result['DP1.00003.001'] == 'Triple aspirated air temperature'

    product_code = 'DP1.00002.001'
    result = act.discovery.get_neon_product_avail(site_code, product_code, print_to_screen=True)
    assert '2017-09' in result
    assert '2022-11' in result

    output_dir = os.path.join(os.getcwd(), site_code + '_' + product_code)
    result = act.discovery.download_neon_data(
        site_code, product_code, '2022-10', output_dir=output_dir
    )
    assert len(result) == 20
    assert any('readme' in r for r in result)
    assert any('sensor_position' in r for r in result)

    result = act.discovery.download_neon_data(
        site_code, product_code, '2022-09', end_date='2022-10', output_dir=output_dir
    )
    assert len(result) == 40
    assert any('readme' in r for r in result)
    assert any('sensor_position' in r for r in result)



"""
Calculate and plot daily daytime temperature averages
-----------------------------------------------------

Example of how to read in MET data and plot up daytime
temperature averages using the add_solar_variable function

Author: Adam Theisen
"""

import matplotlib.pyplot as plt
from arm_test_data import DATASETS

import act

# Read in the sample MET data
met_wildcard_list = [
    'sgpmetE13.b1.20190101.000000.cdf',
    'sgpmetE13.b1.20190102.000000.cdf',
    'sgpmetE13.b1.20190103.000000.cdf',
    'sgpmetE13.b1.20190104.000000.cdf',
    'sgpmetE13.b1.20190105.000000.cdf',
    'sgpmetE13.b1.20190106.000000.cdf',
    'sgpmetE13.b1.20190107.000000.cdf',
]
met_filenames = [DATASETS.fetch(file) for file in met_wildcard_list]
ds = act.io.arm.read_arm_netcdf(met_filenames)

# Add the solar variable, including dawn/dusk to variable
ds = act.utils.geo_utils.add_solar_variable(ds)

# Using the sun variable, only analyze daytime data
ds = ds.where(ds['sun_variable'] == 1)

# Take daily mean using xarray features
ds = ds.resample(time='1d', skipna=True).mean()

# Creat Plot Display
display = act.plotting.TimeSeriesDisplay(ds, figsize=(15, 10))
display.plot('temp_mean', linestyle='solid')
display.day_night_background()
plt.show()

ds.close()



import sys

import numpy as np
import pytest

import act


@pytest.mark.skipif(
    not sys.platform.startswith('linux'),
    reason='To avoid exceeded downloads for testing, only test on linux',
)
def test_get_mplnet_meta():
    output = act.discovery.get_mplnet_meta(
        sites="GSFC", method="data", year="2024", month="09", day="12"
    )

    assert 'id' in output[0]
    assert 'station' in output[0]
    assert output[0]['station']['latitude_unit'] == "deg"

    with pytest.raises(ValueError):
        output = act.discovery.get_mplnet_meta()
    with pytest.raises(ValueError):
        output = act.discovery.get_mplnet_meta(sites=10)


@pytest.mark.skipif(
    not sys.platform.startswith('linux'),
    reason='To avoid exceeded downloads for testing, only test on linux',
)
def test_download_mplnet_data():
    output = act.discovery.download_mplnet_data(
        version=3, level=1, product="NRB", site="GSFC", year="2020", month="09", day="01"
    )

    assert len(output) == 1
    assert output[0][-3:] == "nc4"

    with pytest.raises(ValueError):
        output = act.discovery.download_mplnet_data()
    with pytest.raises(ValueError):
        output = act.discovery.download_mplnet_data(version=3)
    with pytest.raises(ValueError):
        output = act.discovery.download_mplnet_data(version=3, level=1)
    with pytest.raises(ValueError):
        output = act.discovery.download_mplnet_data(version=3, level=1, product='NRB')
    with pytest.raises(ValueError):
        output = act.discovery.download_mplnet_data(version=3, level=1, product='NRB', site="GSFC")
    with pytest.raises(ValueError):
        output = act.discovery.download_mplnet_data(
            version=3, level=1, product='NRB', site="GSFC", year="2020"
        )



"""
Working with embedded quality control variables
-----------------------------------------------

This is an example of how to use existing or create new quality
control varibles. All the tests are located in act/qc/qctests.py
file but called under the qcfilter method.

"""

import numpy as np
from arm_test_data import DATASETS

from act.io.arm import read_arm_netcdf
from act.qc.qcfilter import parse_bit

# Read a data file that does not have any embedded quality control
# variables. This data comes from the example dataset within ACT.
# Can also read data that has existing quality control variables
# and add, manipulate or use those variables the same.
filename_irt = DATASETS.fetch('sgpirt25m20sC1.a0.20190601.000000.cdf')
ds = read_arm_netcdf(filename_irt)

# The name of the data variable we wish to work with
var_name = 'inst_up_long_dome_resist'

# Since there is no embedded quality control varible one will be
# created for us.
# We can start with adding where the data are set to missing value.
# First we will change the first value to NaN to simulate where
# a missing value exist in the data file.
data = ds[var_name].values
data[0] = np.nan
ds[var_name].values = data

# Add a test for where the data are set to missing value.
# Since a quality control variable does not exist in the file
# one will be created as part of adding this test.
result = ds.qcfilter.add_missing_value_test(var_name)

# The returned dictionary will contain the information added to the
# quality control varible for direct use now. Or the information
# can be looked up later for use.
print('\nresult =', result)

# We can add a second test where data is less than a specified value.
result = ds.qcfilter.add_less_test(var_name, 7.8)

# Next we add a test to indicate where a value is greater than
# or equal to a specified number. We also set the assessement
# to a user defined word. The default assessment is "Bad".
result = ds.qcfilter.add_greater_equal_test(var_name, 12, test_assessment='Suspect')

# We can now get the data as a numpy masked array with a mask set
# where the third test we added (greater than or equal to) using
# the result dictionary to get the test number created for us.
data = ds.qcfilter.get_masked_data(var_name, rm_tests=result['test_number'])
print('\nData type =', type(data))

# Or we can get the masked array for all tests that use the assessment
# set to "Bad".
data = ds.qcfilter.get_masked_data(var_name, rm_assessments=['Bad'])

# If we prefer to mask all data for both Bad or Suspect we can list
# as many assessments as needed
data = ds.qcfilter.get_masked_data(var_name, rm_assessments=['Suspect', 'Bad'])
print('\ndata =', data)

# We can convert the masked array into numpy array and choose the fill value.
data = data.filled(fill_value=np.nan)
print('\ndata filled with masked array fill_value =', data)

# We can create our own test by creating an array of indexes of where
# we want the test to be set and call the method to create our own test.
# We can allow the method to pick the test number (next available)
# or set the test number we wan to use. This example uses test number
# 5 to demonstrate how not all tests need to be used in order.
data = ds.qcfilter.get_masked_data(var_name)
diff = np.diff(data)
max_difference = 0.04
data = np.ma.masked_greater(diff, max_difference)
index = np.where(data.mask)[0]
result = ds.qcfilter.add_test(
    var_name,
    index=index,
    test_meaning=f'Difference is greater than {max_difference}',
    test_assessment='Suspect',
    test_number=5,
)

# If we prefer to work with numpy arrays directly we can return the
# data array converted to a numpy array with masked values set
# to NaN. Here we are requesting both Suspect and Bad data be masked.
data = ds.qcfilter.get_masked_data(
    var_name, rm_assessments=['Suspect', 'Bad'], return_nan_array=True
)
print('\nData type =', type(data))
print('data =', data)

# We can see how the quality control data is stored and what assessments,
# or test descriptions are set. Some of the tests have also added attributes to
# store the test limit values.
qc_variable = ds[result['qc_variable_name']]
print('\nQC Variable =', qc_variable)

# The test numbers are not the flag_masks numbers. The flag masks numbers
# are bit-paked numbers used to store what bit is set. To see the test
# numbers we can unpack the bits.
print('\nmask : test')
print('-' * 11)
for mask in qc_variable.attrs['flag_masks']:
    print(mask, ' : ', parse_bit(mask))

# We can also just use the get_masked_data() method to get data
# the same as using ".values" method on the xarray dataset. If we don't
# request any tests or assessments to mask the returned masked array
# will not have any mask set. The returned value is a numpy masked array
# where the raw numpy array is accessable with .data property.
data = ds.qcfilter.get_masked_data(var_name)
print('\nNormal numpy array data values:', data.data)
print('Mask associated with values:', data.mask)

# We can use the get_masked_data() method to return a masked array
# where the test is set in the quality control varialbe, and use the
# masked array method to see if any of the values have the test set.
data = ds.qcfilter.get_masked_data(var_name, rm_tests=3)
print('\nAt least one less than test set =', data.mask.any())
data = ds.qcfilter.get_masked_data(var_name, rm_tests=4)
print('At least one difference test set =', data.mask.any())



"""
Class definitions for cleaning up QC variables to standard
cf-compliance.

"""

import copy
import re

import numpy as np
import xarray as xr

from act.qc.qcfilter import parse_bit


@xr.register_dataset_accessor('clean')
class CleanDataset:
    """
    Class for cleaning up QC variables to standard cf-compliance
    """

    def __init__(self, ds):
        self._ds = ds

    @property
    def matched_qc_variables(self, check_arm_syntax=True):
        """
        Find variables that are QC variables and return list of names.

        Parameters
        ----------
        check_arm_syntax : boolean
            ARM ueses a standard of starting all quality control variables
            with "qc" joined with an underscore. This is a more robust method
            of getting the quality control variables before the standard_name
            attribute is added. If this is true will first check using
            attributes and will then check if variable starts with "qc".

        Returns
        -------
        variables : list of str
            A list of strings containing the name of each variable.

        """

        # Will need to find all historical cases and add to list
        description_list = [
            'See global attributes for individual.+bit descriptions.',
            (
                'This field contains bit packed integer values, where each '
                'bit represents a QC test on the data. Non-zero bits indicate '
                'the QC condition given in the description for those bits; '
                'a value of 0.+ indicates the data has not '
                'failed any QC tests.'
            ),
            (r'This field contains bit packed values which should be ' r'interpreted as listed..+'),
        ]

        # Loop over each variable and look for a match to an attribute that
        # would exist if the variable is a QC variable.
        variables = []
        for var in self._ds.data_vars:
            try:
                if self._ds[var].attrs['standard_name'] == 'quality_flag':
                    variables.append(var)
                    continue
            except KeyError:
                pass

            if check_arm_syntax and var.startswith('qc_'):
                variables.append(var)
                continue

            try:
                for desc in description_list:
                    if re.match(desc, self._ds[var].attrs['description']) is not None:
                        variables.append(var)
                        break
            except KeyError:
                pass

        variables = list(set(variables))

        return variables

    def cleanup(
        self,
        cleanup_arm_qc=True,
        clean_arm_state_vars=None,
        handle_missing_value=True,
        link_qc_variables=True,
        normalize_assessment=False,
        cleanup_cf_qc=True,
        cleanup_incorrect_qc_attributes=True,
        **kwargs,
    ):
        """
        Wrapper method to automatically call all the standard methods
        for dataset cleanup.

        Parameters
        ----------
        cleanup_arm_qc : bool
            Option to clean Xarray dataset from ARM QC to CF QC standards.
        clean_arm_state_vars : list of str
            Option to clean Xarray dataset state variables from ARM to CF
            standards. Pass in list of variable names.
        handle_missing_value : bool
            Go through variables and look for cases where a QC or state varible
            was convereted to a float and missing values set to np.nan. This
            is done because of xarry's default to use mask_and_scale=True.
            This will convert the data type back to integer and replace
            any instances of np.nan to a missing value indicator (most
            likely -9999).
        link_qc_variables : bool
            Option to link QC variablers through ancillary_variables if not
            already set.
        normalize_assessment : bool
            Option to clean up assessments to use the same terminology. Set to
            False for default because should only be an issue after adding DQRs
            and the function to add DQRs calls this method.
        cleanup_incorrect_qc_attributes : bool
            Fix incorrectly named quality control variable attributes before
            converting to standardized QC.
        **kwargs : keywords
            Keyword arguments passed through to clean.clean_arm_qc
            method.

        Examples
        --------
            .. code-block:: python

                files = act.tests.sample_files.EXAMPLE_MET1
                ds = act.io.arm.read_arm_netcdf(files)
                ds.clean.cleanup()

        """
        # There are some QC variables with incorrect bit_#_description attribute names.
        # This will check for the incorrect attribute names and correct to allow next
        # process to work correctly
        if cleanup_incorrect_qc_attributes:
            self._ds.clean.fix_incorrect_variable_bit_description_attributes()

        # Convert ARM QC to be more like CF state fields
        if cleanup_arm_qc:
            self._ds.clean.clean_arm_qc(**kwargs)

        # Convert ARM state fields to be more liek CF state fields
        if clean_arm_state_vars is not None:
            self._ds.clean.clean_arm_state_variables(clean_arm_state_vars)

        # Correctly convert data type because of missing value
        # indicators in state and QC variables. Needs to be run after
        # clean.clean_arm_qc to use CF attribute names.
        if handle_missing_value:
            self._ds.clean.handle_missing_values()

        # Add some ancillary_variables linkages
        # between data variable and QC variable
        if link_qc_variables:
            self._ds.clean.link_variables()

        # Update the terminology used with flag_assessments to be consistent
        if normalize_assessment:
            self._ds.clean.normalize_assessment()

        # Update from CF to standard used in ACT
        if cleanup_cf_qc:
            self._ds.clean.clean_cf_qc(**kwargs)

    def handle_missing_values(self, default_missing_value=np.int32(-9999)):
        """
        Correctly handle missing_value and _FillValue in the dataset.
        xarray will automatically replace missing_value and
        _FillValue in the data with NaN. This is great for data set
        as type float but not great for int data. Can cause issues
        with QC and state fields. This will loop through the array
        looking for state and QC fields and revert them back to int
        data type if upconverted to float to handle NaNs. Issue is that
        xarray will convert data type to float if the attribute is defined
        even if no data are set as missing value. xarray will also then
        remove the missing_value or _FillValue variable attribute. This
        will put the missing_value attribute back if needed.

        Parameters
        ----------
        default_missing_value : numpy int or float
           The default missing value to use if a missing_value attribute
           is not defined but one is needed.

        """
        state_att_names = [
            'flag_values',
            'flag_meanings',
            'flag_masks',
            'flag_attributes',
        ]

        # Look for variables that have 2 of the state_att_names defined
        # as attribures and is of type float. If so assume the variable
        # was incorreclty converted to float type.
        for var in self._ds.data_vars:
            var_att_names = self._ds[var].attrs.keys()
            if len(set(state_att_names) & set(var_att_names)) >= 2 and self._ds[
                var
            ].values.dtype in [
                np.dtype('float16'),
                np.dtype('float32'),
                np.dtype('float64'),
            ]:
                # Look at units variable to see if this is the stupid way some
                # ARM products mix data and state variables. If the units are not
                # in the normal list of unitless type assume this is a data variable
                # and skip. Other option is to lookf or a valid_range attribute
                # and skip. This is commented out for now since the units check
                # appears to be working.
                try:
                    if self._ds[var].attrs['units'] not in ['1', 'unitless', '', ' ']:
                        continue
                except KeyError:
                    pass

                # Change any np.nan values to missing value indicator
                data = self._ds[var].values
                data[np.isnan(data)] = default_missing_value.astype(data.dtype)

                # Convert data to match type of flag_mask or flag_values
                # as the best guess of what type is correct.
                found_dtype = False
                for att_name in ['flag_masks', 'flag_values']:
                    try:
                        att_value = self._ds[var].attrs[att_name]
                        if isinstance(att_value, (list, tuple)):
                            dtype = att_value[0].dtype
                        elif isinstance(att_value, str):
                            dtype = default_missing_value.dtype
                            att_value = att_value.replace(',', ' ').split()
                            att_value = np.array(att_value, dtype=dtype)
                            self._ds[var].attrs[att_name] = att_value
                            dtype = default_missing_value.dtype
                        else:
                            dtype = att_value.dtype
                        data = data.astype(dtype)
                        found_dtype = True
                        break
                    except (KeyError, IndexError, AttributeError):
                        pass

                # If flag_mask or flag_values is not available choose an int type
                # and set data to that type.
                if found_dtype is False:
                    data = data.astype(default_missing_value.dtype)

                # Return data to the dataset and add missing value indicator
                # attribute to variable.
                self._ds[var].values = data
                self._ds[var].attrs['missing_value'] = default_missing_value.astype(data.dtype)

    def get_attr_info(self, variable=None, flag=False):
        """
        Get ARM quality control definitions from the ARM standard
        bit_#_description, ... attributes and return as dictionary.
        Will attempt to guess if the flag is integer or bit packed
        based on what attributes are set.

        Parameters
        ----------
        variable : str
            Variable name to get attribute information. If set to None
            will get global attributes.
        flag : bool
            Optional flag indicating if QC is expected to be bitpacked
            or integer. Flag = True indicates integer QC. Default
            is bitpacked or False.

        Returns
        -------
        attributes dictionary : dict or None
            A dictionary contianing the attribute information converted from
            ARM QC to CF QC. All keys include 'flag_meanings', 'flag_masks',
            'flag_values', 'flag_assessments', 'flag_tests', 'arm_attributes'.
            Returns None if none found.

        """
        string = 'bit'
        if flag:
            string = 'flag'
        else:
            found_string = False
            try:
                if self._ds.attrs['qc_bit_comment']:
                    string = 'bit'
                    found_string = True
            except KeyError:
                pass

            if found_string is False:
                try:
                    if self._ds.attrs['qc_flag_comment']:
                        string = 'flag'
                        found_string = True
                except KeyError:
                    pass

            if found_string is False:
                var = self.matched_qc_variables
                if len(var) > 0:
                    try:
                        if self._ds[variable].attrs['flag_method'] == 'integer':
                            string = 'flag'
                        found_string = True
                        del self._ds[variable].attrs['flag_method']
                    except KeyError:
                        pass

        try:
            if variable:
                attr_description_pattern = r'(^' + string + r')_([0-9]+)_(description$)'
                attr_assessment_pattern = r'(^' + string + r')_([0-9]+)_(assessment$)'
                attr_comment_pattern = r'(^' + string + r')_([0-9]+)_(comment$)'
                attributes = self._ds[variable].attrs
            else:
                attr_description_pattern = r'(^qc_' + string + r')_([0-9]+)_(description$)'
                attr_assessment_pattern = r'(^qc_' + string + r')_([0-9]+)_(assessment$)'
                attr_comment_pattern = r'(^qc_' + string + r')_([0-9]+)_(comment$)'
                attributes = self._ds.attrs
        except KeyError:
            return None

        assessment_bit_num = []
        description_bit_num = []
        comment_bit_num = []
        flag_masks = []
        flag_meanings = []
        flag_assessments = []
        flag_comments = []
        arm_attributes = []

        dtype = np.int32
        for att_name in attributes:
            try:
                description = re.match(attr_description_pattern, att_name)
                description_bit_num.append(int(description.groups()[1]))
                flag_meanings.append(attributes[att_name])
                arm_attributes.append(att_name)
            except AttributeError:
                pass

            try:
                assessment = re.match(attr_assessment_pattern, att_name)
                assessment_bit_num.append(int(assessment.groups()[1]))
                flag_assessments.append(attributes[att_name])
                arm_attributes.append(att_name)
            except AttributeError:
                pass

            try:
                comment = re.match(attr_comment_pattern, att_name)
                comment_bit_num.append(int(comment.groups()[1]))
                flag_comments.append(attributes[att_name])
                arm_attributes.append(att_name)
            except AttributeError:
                pass

        if variable is not None:
            # Try and get the data type from the variable if it is an integer
            # If not an integer make the flag values integers.
            try:
                dtype = self._ds[variable].values.dtype
                if np.issubdtype(dtype, np.integer):
                    pass
                else:
                    dtype = np.int32
            except AttributeError:
                pass

        # Sort on bit number to ensure correct description order
        index = np.argsort(description_bit_num)
        flag_meanings = np.array(flag_meanings)
        description_bit_num = np.array(description_bit_num)
        flag_meanings = flag_meanings[index]
        description_bit_num = description_bit_num[index]

        # Sort on bit number to ensure correct assessment order
        if len(flag_assessments) > 0:
            if len(flag_assessments) < len(flag_meanings):
                for ii in range(1, len(flag_meanings) + 1):
                    if ii not in assessment_bit_num:
                        assessment_bit_num.append(ii)
                        flag_assessments.append('')
            index = np.argsort(assessment_bit_num)
            flag_assessments = np.array(flag_assessments)
            flag_assessments = flag_assessments[index]

        # Sort on bit number to ensure correct comment order
        if len(flag_comments) > 0:
            if len(flag_comments) < len(flag_meanings):
                for ii in range(1, len(flag_meanings) + 1):
                    if ii not in comment_bit_num:
                        comment_bit_num.append(ii)
                        flag_comments.append('')
            index = np.argsort(comment_bit_num)
            flag_comments = np.array(flag_comments)
            flag_comments = flag_comments[index]

        # Convert bit number to mask number
        if len(description_bit_num) > 0:
            flag_masks = np.array(description_bit_num)
            flag_masks = np.left_shift(1, flag_masks - 1)

        # build dictionary to return values
        if len(flag_masks) > 0 or len(description_bit_num) > 0:
            return_dict = dict()
            return_dict['flag_meanings'] = list(np.array(flag_meanings, dtype=str))

            if len(flag_masks) > 0 and max(flag_masks) > np.iinfo(np.uint32).max:
                flag_mask_dtype = np.uint64
            else:
                flag_mask_dtype = np.uint32

            if flag:
                return_dict['flag_values'] = list(np.array(description_bit_num, dtype=dtype))
                return_dict['flag_masks'] = list(np.array([], dtype=flag_mask_dtype))
            else:
                return_dict['flag_values'] = list(np.array([], dtype=dtype))
                return_dict['flag_masks'] = list(np.array(flag_masks, dtype=flag_mask_dtype))

            return_dict['flag_assessments'] = list(np.array(flag_assessments, dtype=str))
            return_dict['flag_tests'] = list(np.array(description_bit_num, dtype=dtype))
            return_dict['flag_comments'] = list(np.array(flag_comments, dtype=str))
            return_dict['arm_attributes'] = arm_attributes

        else:
            # If nothing to return set to None
            return_dict = None

        # If no QC is found but there's a Mentor_QC_Field_Information global attribute,
        # hard code the information.  This is for older ARM files that had QC information
        # in this global attribute.  For these cases, this should hold 100%
        if return_dict is None and 'Mentor_QC_Field_Information' in self._ds.attrs:
            qc_att = self._ds.attrs['Mentor_QC_Field_Information']
            if 'Basic mentor QC checks' in qc_att:
                if len(qc_att) == 920 or len(qc_att) == 1562:
                    return_dict = dict()
                    return_dict['flag_meanings'] = [
                        'Value is equal to missing_value.',
                        'Value is less than the valid_min.',
                        'Value is greater than the valid_max.',
                        'Difference between current and previous values exceeds valid_delta.',
                    ]
                    return_dict['flag_tests'] = [1, 2, 3, 4]
                    return_dict['flag_masks'] = [1, 2, 4, 8]
                    return_dict['flag_assessments'] = ['Bad', 'Bad', 'Bad', 'Indeterminate']
                    return_dict['flag_values'] = []
                    return_dict['flag_comments'] = []
                    return_dict['arm_attributes'] = [
                        'bit_1_description',
                        'bit_1_assessment',
                        'bit_2_description',
                        'bit_2_assessment',
                        'bit_3_description',
                        'bit_3_assessment',
                        'bit_4_description',
                        'bit_4_assessment',
                    ]

        return return_dict

    def clean_arm_state_variables(
        self,
        variables,
        override_cf_flag=True,
        clean_units_string=True,
        integer_flag=True,
        replace_in_flag_meanings=None,
    ):
        """
        Function to clean up state variables to use more CF style.

        Parameters
        ----------
        variables : str or list of str
            List of variable names to update.
        override_cf_flag : bool
            Option to overwrite CF flag_meanings attribute if it exists
            with the values from ARM QC bit_#_description.
        clean_units_string : bool
            Option to update units string if set to 'unitless' to be
            udunits compliant '1'.
        integer_flag : bool
            Pass through keyword of 'flag' for get_attr_info().
        replace_in_flag_meanings : None or string
            Character string to search and replace in each flag meanings array value
            to increase readability since the flag_meanings stored in netCDF file
            is a single character array separated by space character. Alows for
            replacing things like "_" with space character.

        """
        if isinstance(variables, str):
            variables = [variables]

        for var in variables:
            flag_info = self.get_attr_info(variable=var, flag=integer_flag)
            if flag_info is not None:
                # Add new attributes to variable
                for attr in ['flag_values', 'flag_meanings', 'flag_masks']:
                    if len(flag_info[attr]) > 0:
                        # Only add if attribute does not exist.
                        if attr in self._ds[var].attrs.keys() is False:
                            self._ds[var].attrs[attr] = copy.copy(flag_info[attr])
                        # If flag is set, set attribure even if exists
                        elif override_cf_flag:
                            self._ds[var].attrs[attr] = copy.copy(flag_info[attr])

                # Remove replaced attributes
                arm_attributes = flag_info['arm_attributes']
                for attr in arm_attributes:
                    try:
                        del self._ds[var].attrs[attr]
                    except KeyError:
                        pass

            # Check if flag_meanings is string. If so convert to list.
            try:
                flag_meanings = copy.copy(self._ds[var].attrs['flag_meanings'])
                if isinstance(flag_meanings, str):
                    flag_meanings = flag_meanings.split()
                    if replace_in_flag_meanings is not None:
                        for ii, flag_meaning in enumerate(flag_meanings):
                            flag_meaning = flag_meaning.replace(replace_in_flag_meanings, ' ')
                            flag_meanings[ii] = flag_meaning

                    self._ds[var].attrs['flag_meanings'] = flag_meanings
            except KeyError:
                pass

            # Clean up units attribute from unitless to udunits '1'
            try:
                if clean_units_string and self._ds[var].attrs['units'] == 'unitless':
                    self._ds[var].attrs['units'] = '1'
            except KeyError:
                pass

    def correct_valid_minmax(self, qc_variable):
        """
        Function to correct the name and location of quality control limit
        variables that use valid_min and valid_max incorrectly.

        Parameters
        ----------
        qc_variable : str
            Name of quality control variable in the Xarray dataset to correct.

        """
        test_dict = {
            'valid_min': 'fail_min',
            'valid_max': 'fail_max',
            'valid_delta': 'fail_delta',
        }

        aa = re.match(r'^qc_(.+)', qc_variable)
        variable = None
        try:
            variable = aa.groups()[0]
        except AttributeError:
            return

        made_change = False
        try:
            flag_meanings = copy.copy(self._ds[qc_variable].attrs['flag_meanings'])
        except KeyError:
            return

        for attr in test_dict.keys():
            for ii, test in enumerate(flag_meanings):
                if attr in test:
                    flag_meanings[ii] = re.sub(attr, test_dict[attr], test)
                    made_change = True
                    try:
                        self._ds[qc_variable].attrs[test_dict[attr]] = copy.copy(
                            self._ds[variable].attrs[attr]
                        )
                        del self._ds[variable].attrs[attr]
                    except KeyError:
                        pass

        if made_change:
            self._ds[qc_variable].attrs['flag_meanings'] = flag_meanings

    def link_variables(self):
        """
        Add some attributes to link and explain data to QC data relationship.
        """
        for var in self._ds.data_vars:
            aa = re.match(r'^qc_(.+)', var)
            try:
                variable = aa.groups()[0]
                qc_variable = var
            except AttributeError:
                continue

            # Skip data quality fields.
            try:
                if not self._ds[var].attrs["long_name"].startswith("Quality check results on"):
                    continue
            except KeyError:
                pass

            # Get existing data variable ancillary_variables attribute
            try:
                ancillary_variables = self._ds[variable].attrs['ancillary_variables']
            except KeyError:
                ancillary_variables = ''

            # If the QC variable is not in ancillary_variables add
            if qc_variable not in ancillary_variables:
                ancillary_variables = qc_variable

            try:
                self._ds[variable].attrs['ancillary_variables'] = copy.copy(ancillary_variables)
            except KeyError:
                pass

            # Check if QC variable has correct standard_name and iff not fix it.
            correct_standard_name = 'quality_flag'
            try:
                if self._ds[qc_variable].attrs['standard_name'] != correct_standard_name:
                    self._ds[qc_variable].attrs['standard_name'] = correct_standard_name
            except KeyError:
                self._ds[qc_variable].attrs['standard_name'] = correct_standard_name

    def clean_arm_qc(
        self,
        override_cf_flag=True,
        clean_units_string=True,
        correct_valid_min_max=True,
        remove_unset_global_tests=True,
        **kwargs,
    ):
        """
        Method to clean up Xarray dataset QC variables.

        Parameters
        ----------
        override_cf_flag : bool
            Option to overwrite CF flag_masks, flag_meanings, flag_values
            if exists.
        clean_units_string : bool
            Option to clean up units string from 'unitless'
            to udunits compliant '1'.
        correct_valid_min_max : bool
            Option to correct use of valid_min and valid_max with QC variables
            by moving from data variable to QC varible, renaming to fail_min,
            fail_max and fail_detla if the valid_min, valid_max or valid_delta
            is listed in bit discription attribute. If not listed as
            used with QC will assume is being used correctly.
        remove_unset_global_tests : bool
            Option to look for globaly defined tests that are not set at the
            variable level and remove from quality control variable.

        """
        global_qc = self.get_attr_info()
        qc_attributes = None
        for qc_var in self.matched_qc_variables:
            # Clean up units attribute from unitless to udunits '1'
            try:
                if clean_units_string and self._ds[qc_var].attrs['units'] == 'unitless':
                    self._ds[qc_var].attrs['units'] = '1'
            except KeyError:
                pass

            qc_attributes = self.get_attr_info(variable=qc_var)

            if qc_attributes is None:
                qc_attributes = global_qc

            # Add new attributes to variable
            for attr in [
                'flag_masks',
                'flag_meanings',
                'flag_assessments',
                'flag_values',
                'flag_comments',
            ]:
                if qc_attributes is not None and len(qc_attributes[attr]) > 0:
                    # Only add if attribute does not exists
                    if attr in self._ds[qc_var].attrs.keys() is False:
                        self._ds[qc_var].attrs[attr] = copy.copy(qc_attributes[attr])
                    # If flag is set add attribure even if already exists
                    elif override_cf_flag:
                        self._ds[qc_var].attrs[attr] = copy.copy(qc_attributes[attr])

            # Remove replaced attributes
            if qc_attributes is not None:
                arm_attributes = qc_attributes['arm_attributes']
                if 'description' not in arm_attributes:
                    arm_attributes.append('description')
                if 'flag_method' not in arm_attributes:
                    arm_attributes.append('flag_method')
                for attr in arm_attributes:
                    try:
                        del self._ds[qc_var].attrs[attr]
                    except KeyError:
                        pass

            # Check for use of valid_min and valid_max as QC limits and fix
            if correct_valid_min_max:
                self._ds.clean.correct_valid_minmax(qc_var)

        # Clean up global attributes
        if global_qc is not None:
            global_attributes = global_qc['arm_attributes']
            global_attributes.extend(['qc_bit_comment'])
            for attr in global_attributes:
                try:
                    del self._ds.attrs[attr]
                except KeyError:
                    pass

        # If requested remove tests at variable level that were set from global level descriptions.
        # This is assuming the test was only performed if the limit value is listed with the variable
        # even if the global level describes the test.
        if remove_unset_global_tests and global_qc is not None:
            limit_name_list = ['fail_min', 'fail_max', 'fail_delta']

            for qc_var_name in self.matched_qc_variables:
                flag_meanings = self._ds[qc_var_name].attrs['flag_meanings']
                flag_masks = self._ds[qc_var_name].attrs['flag_masks']
                tests_to_remove = []
                for ii, flag_meaning in enumerate(flag_meanings):
                    # Loop over usual test attribute names looking to see if they
                    # are listed in test description. If so use that name for look up.
                    test_attribute_limit_name = None
                    for name in limit_name_list:
                        if name in flag_meaning:
                            test_attribute_limit_name = name
                            break

                    if test_attribute_limit_name is None:
                        continue

                    remove_test = True
                    test_number = parse_bit(flag_masks[ii])[0]
                    for attr_name in self._ds[qc_var_name].attrs:
                        if test_attribute_limit_name == attr_name:
                            remove_test = False
                            break

                        index = self._ds.qcfilter.get_qc_test_mask(
                            qc_var_name=qc_var_name, test_number=test_number
                        )
                        if np.any(index):
                            remove_test = False
                            break

                    if remove_test:
                        tests_to_remove.append(test_number)

                if len(tests_to_remove) > 0:
                    for test_to_remove in tests_to_remove:
                        self._ds.qcfilter.remove_test(
                            qc_var_name=qc_var_name, test_number=test_to_remove
                        )

        # If the QC was not cleaned up because it is not correctly formatted with SERI QC
        # call the SERI QC method.
        if global_qc is None and qc_attributes is None:
            try:
                DQMS = self._ds.attrs['qc_method'] == 'DQMS'
                self._ds.attrs['comment']
            except KeyError:
                try:
                    DQMS = 'sirs_seriqc' in self._ds.attrs['Command_Line']
                except KeyError:
                    DQMS = False

            if DQMS:
                self._ds.clean.clean_seri_qc()

        # If the QC was not cleaned up because it is not correctly formatted with
        # SWATS global attributes call the SWATS QC method.
        try:
            text = 'SWATS QC checks (bit values)'
            SWATS_QC = text in self._ds.attrs['Mentor_QC_Field_Information']
        except KeyError:
            SWATS_QC = False

        if SWATS_QC and global_qc is None and qc_attributes is None:
            self._ds.clean.clean_swats_qc()

    def normalize_assessment(
        self,
        variables=None,
        exclude_variables=None,
        qc_lookup={'Bad': 'Incorrect', 'Indeterminate': 'Suspect'},
    ):
        """
        Method to clean up assessment terms used to be consistent between
        embedded QC and DQRs.

        Parameters
        ----------
        variables : str or list of str
            Optional data variable names to check and normalize. If set to
            None will check all variables.
        exclude_variables : str or list of str
            Optional data variable names to exclude from processing.
        qc_lookup : dict
            Optional dictionary used to convert between terms.

        Examples
        --------
            .. code-block:: python

                ds = act.io.arm.read_arm_netcdf(files)
                ds.clean.normalize_assessment(variables='temp_mean')

            .. code-block:: python

                ds = act.io.arm.read_arm_netcdf(files, cleanup_qc=True)
                ds.clean.normalize_assessment(qc_lookup={'Bad': 'Incorrect', 'Indeterminate': 'Suspect'})

        """

        # Get list of variables if not provided
        if variables is None:
            variables = list(self._ds.data_vars)

        # Ensure variables is a list
        if not isinstance(variables, (list, tuple)):
            variables = [variables]

        # If exclude variables provided remove from variables list
        if exclude_variables is not None:
            if not isinstance(exclude_variables, (list, tuple)):
                exclude_variables = [exclude_variables]

            variables = list(set(variables) - set(exclude_variables))

        # Loop over variables checking if a QC variable exits and use the
        # lookup dictionary to convert the assessment terms.
        for var_name in variables:
            qc_var_name = self._ds.qcfilter.check_for_ancillary_qc(
                var_name, add_if_missing=False, cleanup=False
            )
            if qc_var_name is not None:
                try:
                    flag_assessments = self._ds[qc_var_name].attrs['flag_assessments']
                except KeyError:
                    continue

                for ii, assess in enumerate(flag_assessments):
                    try:
                        flag_assessments[ii] = qc_lookup[assess]
                    except KeyError:
                        continue

    def clean_cf_qc(self, variables=None, sep='__', **kwargs):
        """
        Method to convert the CF standard for QC attributes to match internal
        format expected in the Dataset. CF does not allow string attribute
        arrays, even though netCDF4 does allow string attribute arrays. The quality
        control variables uses and expects lists for flag_meaning, flag_assessments.

        Parameters
        ----------
        variables : str or list of str or None
            Data variable names to convert. If set to None will check all variables.
        sep : str or None
            Separater to use for splitting individual test meanings. Since the CF
            attribute in the netCDF file must be a string and is separated by a
            space character, individual test meanings are connected with a character.
            Default for ACT writing to file is double underscore to preserve underscores
            in variable or attribute names.
        kwargs : dict
            Additional keyword argumnts not used. This is to allow calling multiple
            methods from one method without causing unexpected keyword errors.

        Examples
        --------
            .. code-block:: python

                ds = act.io.arm.read_arm_netcdf(files)
                ds.clean.clean_cf_qc(variables='temp_mean')

            .. code-block:: python

                ds = act.io.arm.read_arm_netcdf(files, cleanup_qc=True)

        """

        # Convert string in to list of string for itteration
        if isinstance(variables, str):
            variables = [variables]

        # If no variables provided, get list of all variables in Dataset
        if variables is None:
            variables = list(self._ds.data_vars)

        for var_name in variables:
            # Check flag_meanings type. If string separate on space character
            # into list. If sep is not None split string on separater to make
            # better looking list of strings.
            try:
                flag_meanings = self._ds[var_name].attrs['flag_meanings']
                if isinstance(flag_meanings, str):
                    flag_meanings = flag_meanings.split()
                    if sep is not None:
                        flag_meanings = [ii.replace(sep, ' ') for ii in flag_meanings]
                    self._ds[var_name].attrs['flag_meanings'] = flag_meanings
            except KeyError:
                pass

            # Check if flag_assessments is a string, split on space character
            # to make list.
            try:
                flag_assessments = self._ds[var_name].attrs['flag_assessments']
                if isinstance(flag_assessments, str):
                    flag_assessments = flag_assessments.split()
                    self._ds[var_name].attrs['flag_assessments'] = flag_assessments
            except KeyError:
                pass

            # Check if flag_masks is a numpy scalar instead of array. If so convert
            # to numpy array. If value is not numpy scalar, turn single value into
            # list.
            try:
                flag_masks = self._ds[var_name].attrs['flag_masks']
                if type(flag_masks).__module__ == 'numpy':
                    if flag_masks.shape == ():
                        self._ds[var_name].attrs['flag_masks'] = np.atleast_1d(flag_masks)

                elif not isinstance(flag_masks, (list, tuple)):
                    self._ds[var_name].attrs['flag_masks'] = [flag_masks]

            except KeyError:
                pass

    def fix_incorrect_variable_bit_description_attributes(self):
        """
        Method to correct incorrectly defined quality control variable attributes.
        There are some datastreams with the attribute names incorrectly having ``qc_``
        prepended to the attribute name. This will fix those attributes so the ``cleanqc``
        method can correctly read the attributes.

        If the variable long_name starts with the string "Quality check results on"
        and a variable attribute follows the pattern ``qc_bit_#_description`` the ``qc_``
        part of the variable attribute will be removed.

        """
        attr_description_pattern = r'^qc_bit_([0-9]+)_description$'
        attr_assessment_pattern = r'^qc_bit_([0-9]+)_assessment$'

        for var_name in self._ds.data_vars:
            try:
                if not self._ds[var_name].attrs['long_name'].startswith("Quality check results on"):
                    continue
            except KeyError:
                continue

            for attr, value in self._ds[var_name].attrs.copy().items():
                for pattern in [attr_description_pattern, attr_assessment_pattern]:
                    description = re.match(pattern, attr)
                    if description is not None:
                        new_attr = attr[3:]
                        self._ds[var_name].attrs[new_attr] = self._ds[var_name].attrs.pop(attr)

    def clean_seri_qc(self):
        """
        Method to apply SERI QC to the quality control variables. The definition of the QC
        is listed in a single global attribute and not easily parsable. This method will update
        the quality control variable to correctly set the test descriptions for each of the
        SERI QC tests defined in the global attributes.

        """
        for var_name in self._ds.data_vars:
            if not self._ds[var_name].attrs['long_name'].startswith("Quality check results on"):
                continue

            qc_var_name = var_name
            var_name = var_name.replace('qc_', '')
            qc_data = self._ds[qc_var_name].values.copy()
            self._ds[qc_var_name] = xr.zeros_like(self._ds[qc_var_name], dtype=np.int32)

            if qc_var_name in [
                "qc_down_short_diffuse",
                "qc_short_direct_normal",
                "qc_down_short_hemisp",
            ]:
                value_number = [1, 2, 3, 6, 7, 8, 9, 94, 95, 96, 97]
                test_number = list(range(2, len(value_number) + 2))
                test_description = [
                    'Passed 1-component test; data fall within max-min limits of Kt,Kn, or Kd',
                    'Passed 2-component test; data fall within 0.03 of the Gompertz boundaries',
                    'Passed 3-component test; data come within +/- 0.03 of satifying Kt=Kn+Kd',
                    'Value estimated; passes all pertinent SERI QC tests',
                    'Failed 1-component test; lower than allowed minimum',
                    'Falied 1-component test; higher than allowed maximum',
                    'Passed 3-component test but failed 2-component test by >0.05',
                    'Data fall into a physically impossible region where Kn>Kt by K-space distances of 0.05 to 0.10.',
                    'Data fall into a physically impossible region where Kn>Kt by K-space distances of 0.10 to 0.15.',
                    'Data fall into a physically impossible region where Kn>Kt by K-space distances of 0.15 to 0.20.',
                    'Data fall into a physically impossible region where Kn>Kt by K-space distances of >= 0.20.',
                ]
                test_assessment = [
                    'Not failing',
                    'Not failing',
                    'Not failing',
                    'Not failing',
                    'Bad',
                    'Bad',
                    'Indeterminate',
                    'Bad',
                    'Bad',
                    'Bad',
                    'Bad',
                ]
            elif qc_var_name in ["qc_up_long_hemisp", "qc_down_long_hemisp_shaded"]:
                value_number = [1, 2, 7, 8, 31]
                test_number = list(range(2, len(value_number) + 2))
                test_description = [
                    'Passed 1-component test; data fall within max-min limits of up_long_hemisp and down_long_hemisp_shaded, but short_direct_normal and down_short_hemisp or down_short_diffuse fail the SERI QC tests.',
                    'Passed 2-component test; data fall within max-min limits of up_long_hemisp and down_long_hemisp_shaded, and short_direct_normal, or down_short_hemisp and down_short_diffuse pass the SERI QC tests while the difference between down_short_hemisp and down_short_diffuse is greater than 20 W/m2.',
                    'Failed 1-component test; lower than allowed minimum',
                    'Failed 1-component test; higher than allowed maximum',
                    'Failed 2-component test',
                ]
                test_assessment = [
                    'Not failing',
                    'Not failing',
                    'Bad',
                    'Bad',
                    'Bad',
                ]
            elif qc_var_name in ["qc_up_short_hemisp"]:
                value_number = [1, 2, 7, 8, 31]
                test_number = list(range(2, len(value_number) + 2))
                test_description = [
                    'Passed 1-component test',
                    'Passed 2-component test',
                    'Failed 1-component test; lower than allowed minimum',
                    'Failed 1-component test; higher than allowed maximum',
                    'Failed 2-component test; solar zenith angle is less than 80 degrees and down_short_hemisp is 0 or missing',
                ]
                test_assessment = [
                    'Not failing',
                    'Not failing',
                    'Bad',
                    'Bad',
                    'Bad',
                ]

            self._ds[var_name].attrs['ancillary_variables'] = qc_var_name
            self._ds[qc_var_name].attrs['standard_name'] = 'quality_flag'
            self._ds[qc_var_name].attrs['flag_masks'] = []
            self._ds[qc_var_name].attrs['flag_meanings'] = []
            self._ds[qc_var_name].attrs['flag_assessments'] = []

            self._ds.qcfilter.add_missing_value_test(var_name)

            for ii, _ in enumerate(value_number):
                index = qc_data == value_number[ii]
                self._ds.qcfilter.add_test(
                    var_name,
                    index=index,
                    test_number=test_number[ii],
                    test_meaning=test_description[ii],
                    test_assessment=test_assessment[ii],
                )

            if qc_var_name in [
                "qc_down_short_diffuse",
                "qc_short_direct_normal",
                "qc_down_short_hemisp",
            ]:
                calculation = ((qc_data + 2) / 4.0) % 4
                calculation = calculation.astype(np.int16)
                value_number = [0, 1, 2, 3]
                test_description = [
                    'Parameter too low by 3-component test (Kt=Kn+Kd)',
                    'Parameter too high by 3-component test (Kt=Kn+Kd)',
                    'Parameter too low by 2-component test (Gompertz boundary)',
                    'Parameter too high by 2-component test (Gompertz boundary)',
                ]
                test_assessment = ['Bad', 'Bad', 'Bad', 'Bad']
                for ii, _ in enumerate(value_number):
                    index = (qc_data >= 10) & (qc_data <= 93) & (calculation == value_number[ii])
                    self._ds.qcfilter.add_test(
                        var_name,
                        index=index,
                        test_meaning=test_description[ii],
                        test_assessment=test_assessment[ii],
                    )

    def clean_swats_qc(self, fix_data_units=True):
        """
        Method to apply SWATS global attribute quality control definition to the
        quality control variables.

        Parameters
        ----------
        fix_data_units : bool
            The units string for some data variables incorrectly defines degrees Celsius
            as 'C' insted of the udunits 'degC'. When set to true those units strings
            are updated.

        """

        for var_name in self._ds.data_vars:
            if fix_data_units:
                try:
                    unit = self._ds[var_name].attrs['units']
                    if unit == 'C':
                        self._ds[var_name].attrs['units'] = 'degC'
                except KeyError:
                    pass

            if not self._ds[var_name].attrs['long_name'].startswith("Quality check results on"):
                continue

            qc_var_name = var_name
            self._ds[qc_var_name].attrs['flag_masks'] = [1, 2, 4, 8]
            self._ds[qc_var_name].attrs['flag_meanings'] = [
                'Value is set to missing_value.',
                'Data value less than valid_min.',
                'Data value greater than valid_max.',
                'Difference between current and previous values exceeds valid_delta.',
            ]
            self._ds[qc_var_name].attrs['flag_assessments'] = [
                'Bad',
                'Bad',
                'Bad',
                'Indeterminate',
            ]

            self._ds.clean.correct_valid_minmax(qc_var_name)

        del self._ds.attrs['Mentor_QC_Field_Information']



import pytest

import act


def test_get_improve():
    ds = act.discovery.get_improve_data(site_id='244', start_date='1/1/2023', end_date='12/31/2023')

    assert len(list(ds)) == 216
    assert 'lat' in ds
    assert 'lon' in ds
    assert len(ds.time.values) == 121
    assert 'aluminum_fine' in ds
    assert ds['ammonium_nitrate_fine'].values[0] == 1.41363

    with pytest.raises(ValueError):
        ds = act.discovery.get_improve_data()
    with pytest.raises(ValueError):
        ds = act.discovery.get_improve_data(site_id='244')
    with pytest.raises(ValueError):
        ds = act.discovery.get_improve_data(site_id='244', start_date='1/1/2023')



import copy

import numpy as np

from act.io.arm import read_arm_netcdf
from act.tests import EXAMPLE_MET1


def test_compare_time_series_trends():
    drop_vars = [
        'base_time',
        'time_offset',
        'atmos_pressure',
        'qc_atmos_pressure',
        'temp_std',
        'rh_mean',
        'qc_rh_mean',
        'rh_std',
        'vapor_pressure_mean',
        'qc_vapor_pressure_mean',
        'vapor_pressure_std',
        'wspd_arith_mean',
        'qc_wspd_arith_mean',
        'wspd_vec_mean',
        'qc_wspd_vec_mean',
        'wdir_vec_mean',
        'qc_wdir_vec_mean',
        'wdir_vec_std',
        'tbrg_precip_total',
        'qc_tbrg_precip_total',
        'tbrg_precip_total_corr',
        'qc_tbrg_precip_total_corr',
        'org_precip_rate_mean',
        'qc_org_precip_rate_mean',
        'pwd_err_code',
        'pwd_mean_vis_1min',
        'qc_pwd_mean_vis_1min',
        'pwd_mean_vis_10min',
        'qc_pwd_mean_vis_10min',
        'pwd_pw_code_inst',
        'qc_pwd_pw_code_inst',
        'pwd_pw_code_15min',
        'qc_pwd_pw_code_15min',
        'pwd_pw_code_1hr',
        'qc_pwd_pw_code_1hr',
        'pwd_precip_rate_mean_1min',
        'qc_pwd_precip_rate_mean_1min',
        'pwd_cumul_rain',
        'qc_pwd_cumul_rain',
        'pwd_cumul_snow',
        'qc_pwd_cumul_snow',
        'logger_volt',
        'qc_logger_volt',
        'logger_temp',
        'qc_logger_temp',
        'lat',
        'lon',
        'alt',
    ]
    ds = read_arm_netcdf(EXAMPLE_MET1, drop_variables=drop_vars)
    ds.clean.cleanup()
    ds2 = copy.deepcopy(ds)

    var_name = 'temp_mean'
    qc_var_name = ds.qcfilter.check_for_ancillary_qc(
        var_name, add_if_missing=False, cleanup=False, flag_type=False
    )
    ds.qcfilter.compare_time_series_trends(
        var_name=var_name,
        time_shift=60,
        comp_var_name=var_name,
        comp_dataset=ds2,
        time_qc_threshold=60 * 10,
    )

    test_description = (
        'Time shift detected with Minimum Difference test. Comparison of '
        'temp_mean with temp_mean off by 0 seconds exceeding absolute '
        'threshold of 600 seconds.'
    )
    assert ds[qc_var_name].attrs['flag_meanings'][-1] == test_description

    time = ds2['time'].values + np.timedelta64(1, 'h')
    time_attrs = ds2['time'].attrs
    ds2 = ds2.assign_coords({'time': time})
    ds2['time'].attrs = time_attrs

    ds.qcfilter.compare_time_series_trends(
        var_name=var_name, comp_dataset=ds2, time_step=60, time_match_threshhold=50
    )

    test_description = (
        'Time shift detected with Minimum Difference test. Comparison of '
        'temp_mean with temp_mean off by 3600 seconds exceeding absolute '
        'threshold of 900 seconds.'
    )
    assert ds[qc_var_name].attrs['flag_meanings'][-1] == test_description



"""
Module containing utilities for the data.

"""

import importlib
import json
import re
import warnings
from os import PathLike
from pathlib import Path

import metpy
import numpy as np
import pint
import requests
import xarray as xr

spec = importlib.util.find_spec('pyart')
if spec is not None:
    PYART_AVAILABLE = True
else:
    PYART_AVAILABLE = False


@xr.register_dataset_accessor('utils')
class ChangeUnits:
    """
    Class for updating units in the dataset. Data values and units attribute
    are updated in place. Coordinate variables can not be updated in place. Must
    use new returned dataset when updating coordinage varibles.
    """

    def __init__(self, ds):
        self._ds = ds

    def change_units(
        self,
        variables=None,
        desired_unit=None,
        skip_variables=None,
        skip_standard=True,
        verbose=False,
        raise_error=False,
    ):
        """
        Parameters
        ----------
        variables : None, str or list of str
            Variable names to attempt to change units.
        desired_unit : str
            Desired udunits unit string.
        skip_variables : None, str or list of str
            Variable names to skip. Works well when not providing a variables
            keyword.
        skip_standard : boolean
            Flag indicating the QC variables that will not need changing are
            skipped. Makes the processing faster when processing all variables
            in dataset.
        verbose : boolean
            Option to print statement when an attempted conversion fails. Set to False
            as default because many units strings are not udunits complient and when
            trying to convert all varialbes of a type of units (eg temperature) the code
            can print a lot of unecessary information.
        raise_error : boolean
            Raise an error if conversion is not successful.

        Returns
        -------
        dataset : xarray.dataset
            A new dataset if the coordinate variables are updated. Required to
            use returned dataset if coordinage variabels are updated,
            otherwise the dataset is updated in place.
        """

        if variables is not None and isinstance(variables, str):
            variables = [variables]

        if skip_variables is not None and isinstance(skip_variables, str):
            skip_variables = [skip_variables]

        if desired_unit is None:
            raise ValueError("Need to provide 'desired_unit' keyword for .change_units() method")

        if variables is None:
            variables = list(self._ds.data_vars)

        if skip_variables is not None:
            variables = list(set(variables) - set(skip_variables))

        for var_name in variables:
            try:
                if self._ds[var_name].attrs['standard_name'] == 'quality_flag':
                    continue
            except KeyError:
                pass

            try:
                data = convert_units(
                    self._ds[var_name].values,
                    self._ds[var_name].attrs['units'],
                    desired_unit,
                )
                try:
                    self._ds[var_name].values = data
                    self._ds[var_name].attrs['units'] = desired_unit
                except ValueError:
                    attrs = self._ds[var_name].attrs
                    self._ds = self._ds.assign_coords({var_name: data})
                    attrs['units'] = desired_unit
                    self._ds[var_name].attrs = attrs
            except (
                KeyError,
                TypeError,
                pint.errors.DimensionalityError,
                pint.errors.UndefinedUnitError,
                np._core._exceptions.UFuncTypeError,
            ):
                if raise_error:
                    raise ValueError(
                        f"Unable to convert '{var_name}' to units of '{desired_unit}'."
                    )
                elif verbose:
                    print(
                        f"\n    Unable to convert '{var_name}' to units of '{desired_unit}'. "
                        f"Skipping unit converstion for '{var_name}'.\n"
                    )

        return self._ds


# @xr.register_dataset_accessor('utils')
class DatastreamParserARM:
    '''
    Class to parse ARM datastream names or filenames into its components.
    Will return None for each attribute if not extracted from the filename.

    Attributes
    ----------
    site : str or None
        The site code extracted from the filename.
    datastream_class : str
        The datastream class extracted from the filename.
    facility : str or None
        The datastream facility code extracted from the filename.
    level : str or None
        The datastream level code extracted from the filename.
    datastream : str or None
        The datastram extracted from the filename.
    date : str or None
        The date extracted from the filename.
    time : str or None
        The time extracted from the filename.
    ext : str or None
        The file extension extracted from the filename.

    Example
    -------
    >>> from act.utils.data_utils import DatastreamParserARM
    >>> file = 'sgpmetE13.b1.20190501.024254.nc'
    >>> fn_obj = DatastreamParserARM(file)
    >>> fn_obj.site
    'sgp'
    >>> fn_obj.datastream_class
    'met'


    '''

    def __init__(self, ds=''):
        '''
        Constructor that initializes datastream data member and runs
        parse_datastream class method.  Also converts datastream name to
        lower case before parsing.

        ds : str
            The datastream or filename to parse

        '''
        if isinstance(ds, (str, PathLike)):
            self.__datastream = Path(ds).name
        else:
            raise ValueError('Datastream or filename name must be a string or pathlib.PosixPath.')

        try:
            self.__parse_datastream()
        except ValueError:
            self.__site = None
            self.__class = None
            self.__facility = None
            self.__datastream = None
            self.__level = None
            self.__date = None
            self.__time = None
            self.__ext = None

    def __parse_datastream(self):
        '''
        Private method to parse datastream name into its various components
        (site, class, facility, and data level.  Is called automatically by
        constructor when object of class is instantiated and when the
        set_datastream method is called to reset the object.

        '''
        # Import the built-in match function from regular expression library
        # self.__datastream = self.__datastream
        tempstring = self.__datastream.split('.')

        # Check to see if ARM-standard filename was passed
        self.__ext = None
        self.__time = None
        self.__date = None
        self.__level = None
        self.__site = None
        self.__class = None
        self.__facility = None
        if len(tempstring) >= 5:
            self.__ext = tempstring[4]

        if len(tempstring) >= 4:
            self.__time = tempstring[3]

        if len(tempstring) >= 3:
            self.__date = tempstring[2]

        if len(tempstring) >= 2:
            m = re.match('[abcs0][0123456789]', tempstring[1])
            if m is not None:
                self.__level = m.group()

        match = False
        m = re.search(r'(^[a-z]{3})(\w+)([A-Z]{1}\d{1,2})$', tempstring[0])
        if m is not None:
            self.__site = m.group(1)
            self.__class = m.group(2)
            self.__facility = m.group(3)
            match = True

        if not match:
            m = re.search(r'(^[a-z]{3})([^A-Z]+)$', tempstring[0])
            if m is not None:
                self.__site = m.group(1)
                self.__class = m.group(2)
                match = True

        if not match and len(tempstring[0]) == 3:
            m = re.search(r'(^[a-z]{3})', tempstring[0])
            if m is not None:
                self.__site = m.group(1)
                match = True

        if not match:
            raise ValueError(self.__datastream)

    def set_datastream(self, ds):
        '''
        Method used to set or reset object by passing a new datastream name.

        '''

        self.__init__(ds)

    @property
    def datastream(self):
        '''
        Property returning current datastream name stored in object in
        standard lower case.  Will return the datastrem with no level if
        unavailable.

        '''

        try:
            return ''.join((self.__site, self.__class, self.__facility, '.', self.__level))
        except TypeError:
            return None

    @property
    def site(self):
        '''
        Property returning current site name stored in object in standard
        lower case.

        '''

        return self.__site

    @property
    def datastream_class(self):
        '''
        Property returning current datastream class name stored in object in
        standard lower case.  Could not use class as attribute name since it
        is a reserved word in Python

        '''

        return self.__class

    @property
    def facility(self):
        '''
        Property returning current facility name stored in object in
        standard upper case.

        '''

        try:
            return self.__facility.upper()
        except AttributeError:
            return self.__facility

    @property
    def level(self):
        '''
        Property returning current data level stored in object in standard
        lower case.
        '''

        return self.__level

    @property
    def datastream_standard(self):
        '''
        Property returning datastream name in ARM-standard format with
        facility in caps.  Will return the datastream name with no level if
        unavailable.
        '''

        try:
            return ''.join((self.site, self.datastream_class, self.facility, '.', self.level))

        except TypeError:
            return None

    @property
    def date(self):
        '''
        Property returning date from filename.
        '''

        return self.__date

    @property
    def time(self):
        '''
        Property returning time from filename.
        '''

        return self.__time

    @property
    def ext(self):
        '''
        Property returning file extension from filename.
        '''

        return self.__ext


def assign_coordinates(ds, coord_list):
    """
    This procedure will create a new ACT dataset whose coordinates are
    designated to be the variables in a given list. This helps make data
    slicing via xarray and visualization easier.

    Parameters
    ----------
    ds : ACT Dataset
        The ACT Dataset to modify the coordinates of.
    coord_list : dict
        The list of variables to assign as coordinates, given as a dictionary
        whose keys are the variable name and values are the dimension name.

    Returns
    -------
    new_ds : ACT Dataset
        The new ACT Dataset with the coordinates assigned to be the given
        variables.

    """
    # Check to make sure that user assigned valid entries for coordinates

    for coord in coord_list.keys():
        if coord not in ds.variables.keys():
            raise KeyError(coord + ' is not a variable in the Dataset.')

        if ds.sizes[coord_list[coord]] != len(ds.variables[coord]):
            raise IndexError(
                coord + ' must have the same ' + 'value as length of ' + coord_list[coord]
            )

    new_ds_dict = {}
    for variable in ds.variables.keys():
        my_coord_dict = {}
        dataarray = ds[variable]
        if len(dataarray.dims) > 0:
            for coord in coord_list.keys():
                if coord_list[coord] in dataarray.dims:
                    my_coord_dict[coord_list[coord]] = ds[coord]

        if variable not in my_coord_dict.keys() and variable not in ds.dims:
            the_dataarray = xr.DataArray(dataarray.data, coords=my_coord_dict, dims=dataarray.dims)
            new_ds_dict[variable] = the_dataarray

    new_ds = xr.Dataset(new_ds_dict, coords=my_coord_dict)

    return new_ds


def add_in_nan(time, data):
    """
    This procedure adds in NaNs when there is a larger than expected time step.
    This is useful for timeseries where there is a gap in data and need a
    NaN value to stop plotting from connecting data over the large data gap.

    Parameters
    ----------
    time : 1D array of numpy datetime64 or Xarray DataArray of datetime64
        Times in the timeseries.
    data : 1D or 2D numpy array or Xarray DataArray
        Array containing the data. The 0 axis corresponds to time.

    Returns
    -------
    time : numpy array or Xarray DataArray
        The array containing the new times including a NaN filled
        sampe or slice if multi-dimensional.
        The intervals are determined by the mode of the timestep in *time*.
    data : numpy array or Xarray DataArray
        The array containing the NaN-indserted data.

    """

    time_is_DataArray = False
    data_is_DataArray = False
    if isinstance(time, xr.core.dataarray.DataArray):
        time_is_DataArray = True
        time_attributes = time.attrs
        time_dims = time.dims
    if isinstance(data, xr.core.dataarray.DataArray):
        data_is_DataArray = True
        data_attributes = data.attrs
        data_dims = data.dims

    # Return if time dimension is only size one since we can't do differences.
    if time.size > 2:
        data = np.asarray(data)
        time = np.asarray(time)
        # Not sure if we need to set to second data type to make it work better.
        # Leaving code in here in case we need to update.
        # diff = np.diff(time.astype('datetime64[s]'), 1)
        diff = np.diff(time, 1)

        # Use np.unique instead of stats.mode for timedelta arrays (scipy 1.11+ compatibility)
        unique_vals, counts = np.unique(diff, return_counts=True)
        mode = unique_vals[np.argmax(counts)]
        index = np.where(diff > (2.0 * mode))

        # If the data is not float time and we try to insert a NaN it will
        # not auto upconvert the data. Need to convert before inserting NaN.
        if len(index) > 0 and np.issubdtype(data.dtype, np.integer):
            data = data.astype('float32')

        offset = 0
        for i in index[0]:
            corr_i = i + offset

            if len(data.shape) == 1:
                # For line plotting adding a NaN will stop the connection of the line
                # between points. So we just need to add a NaN anywhere between the points.
                corr_i = i + offset
                time_added = time[corr_i] + (time[corr_i + 1] - time[corr_i]) / 2.0
                time = np.insert(time, corr_i + 1, time_added)
                data = np.insert(data, corr_i + 1, np.nan, axis=0)
                offset += 1
            else:
                # For 2D plots need to add a NaN right after and right before the data
                # to correctly mitigate streaking with pcolormesh.
                time_added_1 = time[corr_i] + 1  # One time step after
                time_added_2 = time[corr_i + 1] - 1  # One time step before
                time = np.insert(time, corr_i + 1, [time_added_1, time_added_2])
                data = np.insert(data, corr_i + 1, np.nan, axis=0)
                data = np.insert(data, corr_i + 2, np.nan, axis=0)
                offset += 2

        if time_is_DataArray:
            time = xr.DataArray(time, attrs=time_attributes, dims=time_dims)

        if data_is_DataArray:
            data = xr.DataArray(data, attrs=data_attributes, dims=data_dims)

    return time, data


def get_missing_value(
    ds,
    variable,
    default=-9999,
    add_if_missing_in_ds=False,
    use_FillValue=False,
    nodefault=False,
):
    """
    Function to get missing value from missing_value or _FillValue attribute.
    Works well with catching errors and allows for a default value when a
    missing value is not listed in the dataset. You may get strange results
    becaus xarray will automatically convert all missing_value or
    _FillValue to NaN and then remove the missing_value and
    _FillValue variable attribute when reading data with default settings.

    Parameters
    ----------
    ds : xarray.Dataset
        Xarray dataset containing data variable.
    variable : str
        Variable name to use for getting missing value.
    default : int or float
        Default value to use if missing value attribute is not in dataset.
    add_if_missing_in_ds : bool
        Boolean to add to the dataset if does not exist. Default is False.
    use_FillValue : bool
        Boolean to use _FillValue instead of missing_value. If missing_value
        does exist and _FillValue does not will add _FillValue
        set to missing_value value.
    nodefault : bool
        Option to use this to check if the varible has a missing value set and
        do not want to get default as return. If the missing value is found
        will return, else will return None.

    Returns
    -------
    missing : scalar int or float (or None)
        Value used to indicate missing value matching type of data or None if
        nodefault keyword set to True.

    Examples
    --------
    .. code-block:: python

        from act.utils import get_missing_value

        missing = get_missing_value(dq_ds, "temp_mean")
        print(missing)
        -9999.0

    """
    in_ds = False
    if use_FillValue:
        missing_atts = ['_FillValue', 'missing_value']
    else:
        missing_atts = ['missing_value', '_FillValue']

    for att in missing_atts:
        try:
            missing = ds[variable].attrs[att]
            in_ds = True
            break
        except (AttributeError, KeyError):
            missing = default

    # Check if do not want a default value retured and a value
    # was not fund.
    if nodefault is True and in_ds is False:
        missing = None
        return missing

    # Check data type and try to match missing_value to the data type of data
    try:
        missing = ds[variable].data.dtype.type(missing)
    except KeyError:
        pass
    except AttributeError:
        print(
            ('--- AttributeError: Issue trying to get data type ' + 'from "{}" data ---').format(
                variable
            )
        )

    # If requested add missing value to the dataset
    if add_if_missing_in_ds and not in_ds:
        try:
            ds[variable].attrs[missing_atts[0]] = missing
        except KeyError:
            print(
                ('---  KeyError: Issue trying to add "{}" ' + 'attribute to "{}" ---').format(
                    missing_atts[0], variable
                )
            )

    return missing


def convert_units(data, in_units, out_units):
    """
    Wrapper function around library to convert data using unit strings.
    Currently using pint units library. Will attempt to preserve numpy
    data type, but will upconvert to numpy float64 if need to change
    data type for converted values.

    Parameters
    ----------
    data : list, tuple or numpy array
        Data array to be modified.
    in_units : str
        Units scalar string of input data array.
    out_units : str
        Units scalar string of desired output data array.

    Returns
    -------
    data : numpy array
        Data array converted into new units.

    Examples
    --------
    > data = np.array([1,2,3,4,5,6])
    > data = convert_units(data, 'cm', 'm')
    > data
    array([0.01, 0.02, 0.03, 0.04, 0.05, 0.06])


    """
    # Fix historical and current incorrect usage of units.
    convert_dict = {
        'C': 'degC',
        'F': 'degF',
        '%': 'percent',  # Pint does not like this symbol with .to('%')
        '1': 'unitless',  # Pint does not like a number
    }

    if in_units in convert_dict:
        in_units = convert_dict[in_units]

    if out_units in convert_dict:
        out_units = convert_dict[out_units]

    if in_units == out_units:
        return data

    # Instantiate the registry
    ureg = pint.UnitRegistry(autoconvert_offset_to_baseunit=True)

    # Add missing units and conversions
    ureg.define('fraction = []')
    ureg.define('unitless = []')

    if not isinstance(data, np.ndarray):
        data = np.array(data)

    data_type = data.dtype
    data_type_kind = data.dtype.kind

    # Do the conversion magic
    data = (data * ureg(in_units)).to(out_units)
    data = data.magnitude

    # The data type may be changed by pint. This is a side effect
    # of pint changing the datatype to float. Check if the converted values
    # need float precision. If so leave, if not change back to orginal
    # precision after checking if the precsion is not lost with the orginal
    # data type.
    if (
        data_type_kind == 'i'
        and np.nanmin(data) >= np.iinfo(data_type).min
        and np.nanmax(data) <= np.iinfo(data_type).max
        and np.all(np.mod(data, 1) == 0)
    ):
        data = data.astype(data_type)

    return data


def ts_weighted_average(ts_dict):
    """
    Program to take in multiple difference time-series and average them
    using the weights provided. This assumes that the variables passed in
    all have the same units. Please see example gallery for an example.

    NOTE: All weights should add up to 1

    Parameters
    ----------
    ts_dict : dict
        Dictionary containing datastream, variable, weight, and datasets

        .. code-block:: python

            t_dict = {
                "sgpvdisC1.b1": {
                    "variable": "rain_rate",
                    "weight": 0.05,
                    "ds": ds,
                },
                "sgpmetE13.b1": {
                    "variable": [
                        "tbrg_precip_total",
                        "org_precip_rate_mean",
                        "pwd_precip_rate_mean_1min",
                    ],
                    "weight": [0.25, 0.05, 0.0125],
                },
            }

    Returns
    -------
    data : numpy array
        Variable of time-series averaged data

    """

    # Run through each datastream/variable and get data
    da_array = []
    data = 0.0
    for d in ts_dict:
        for i, v in enumerate(ts_dict[d]['variable']):
            new_name = '_'.join([d, v])
            # Since many variables may have same name, rename with datastream
            da = ts_dict[d]['ds'][v].rename(new_name)

            # Apply Weights to Data
            da.values = da.values * ts_dict[d]['weight'][i]
            da_array.append(da)

    da = xr.merge(da_array)

    # Stack all the data into a 2D time series
    data = None
    for i, d in enumerate(da):
        if i == 0:
            data = da[d].values
        else:
            data = np.vstack((data, da[d].values))

    # Sum data across each time sample
    data = np.nansum(data, 0)

    # Add data to data array and return
    dims = ts_dict[list(ts_dict.keys())[0]]['ds'].dims
    da_xr = xr.DataArray(
        data,
        dims=dims,
        coords={'time': ts_dict[list(ts_dict.keys())[0]]['ds']['time']},
    )
    da_xr.attrs['long_name'] = 'Weighted average of ' + ', '.join(list(ts_dict.keys()))

    return da_xr


def accumulate_precip(ds, variable, time_delta=None):
    """
    Program to accumulate rain rates from an act xarray dataset and insert
    variable back into an act xarray dataset with "_accumulated" appended to
    the variable name. Please verify that your units are accurately described
    in the data.

    Parameters
    ----------
    ds : xarray.DataSet
        ACT Xarray dataset.
    variable : string
        Variable name.
    time_delta : float
        Time delta to caculate precip accumulations over.
        Useful if full time series is not passed in.

    Returns
    -------
    ds : xarray.DataSet
        ACT Xarray dataset with variable_accumulated.

    """
    # Get Data, time, and metadat
    data = ds[variable]
    time = ds.coords['time']
    units = ds[variable].attrs['units']

    # Calculate mode of the time samples(i.e. 1 min vs 1 sec)
    if time_delta is None:
        diff = np.diff(time.values, 1) / np.timedelta64(1, 's')
        # Use np.unique instead of stats.mode (scipy 1.11+ compatibility)
        unique_vals, counts = np.unique(diff, return_counts=True)
        t_delta = unique_vals[np.argmax(counts)]
    else:
        t_delta = time_delta

    # Calculate the accumulation based on the units
    t_factor = t_delta / 60.0
    if units == 'mm/hr':
        data = data * (t_factor / 60.0)

    accum = np.nancumsum(data.values)

    # Add accumulated variable back to the dataset
    long_name = 'Accumulated precipitation'
    attrs = {'long_name': long_name, 'units': 'mm'}
    ds['_'.join([variable, 'accumulated'])] = xr.DataArray(
        accum, coords=ds[variable].coords, attrs=attrs
    )

    return ds


def create_pyart_obj(
    ds,
    variables=None,
    sweep=None,
    azimuth=None,
    elevation=None,
    range_var=None,
    sweep_start=None,
    sweep_end=None,
    lat=None,
    lon=None,
    alt=None,
    sweep_mode='ppi',
    sweep_az_thresh=10.0,
    sweep_el_thresh=0.5,
):
    """
    Produces a Py-ART radar object based on data in the ACT Xarray dataset.

    Parameters
    ----------
    ds : xarray.DataSet
        ACT Xarray dataset.
    variables : list
        List of variables to add to the radar object, will default to all
        variables.
    sweep : string
        Name of variable that has sweep information. If none, will try and
        calculate from the azimuth and elevation.
    azimuth : string
        Name of azimuth variable. Will try and find one if none given.
    elevation : string
        Name of elevation variable. Will try and find one if none given.
    range_var : string
        Name of the range variable. Will try and find one if none given.
    sweep_start : string
        Name of variable with sweep start indices.
    sweep_end : string
        Name of variable with sweep end indices.
    lat : string
        Name of latitude variable. Will try and find one if none given.
    lon : string
        Name of longitude variable. Will try and find one if none given.
    alt : string
        Name of altitude variable. Will try and find one if none given.
    sweep_mode : string
        Type of scan. Defaults to PPI.
    sweep_az_thresh : float
        If calculating sweep numbers, the maximum change in azimuth before new
        sweep.
    sweep_el_thresh : float
        If calculating sweep numbers, the maximum change in elevation before
        new sweep.

    Returns
    -------
    radar : radar.Radar
        Py-ART Radar Object.

    """

    if not PYART_AVAILABLE:
        raise ImportError(
            'Py-ART needs to be installed on your system to convert to ' 'Py-ART Object.'
        )
    else:
        import pyart
    # Get list of variables if none provided
    if variables is None:
        variables = list(ds.keys())

    # Determine the sweeps if not already in a variable$a
    if sweep is None:
        swp = np.zeros(ds.sizes['time'])
        for key in ds.variables.keys():
            if len(ds.variables[key].shape) == 2:
                total_rays = ds.variables[key].shape[0]
                break
        nsweeps = int(total_rays / ds.variables['time'].shape[0])
    else:
        swp = ds[sweep].values
        nsweeps = ds[sweep].values

    # Get coordinate variables
    if lat is None:
        lat = [s for s in variables if 'latitude' in s]
        if len(lat) == 0:
            lat = [s for s in variables if 'lat' in s]
        if len(lat) == 0:
            raise ValueError(
                'Latitude variable not set and could not be ' 'discerned from the data.'
            )
        else:
            lat = lat[0]

    if lon is None:
        lon = [s for s in variables if 'longitude' in s]
        if len(lon) == 0:
            lon = [s for s in variables if 'lon' in s]
        if len(lon) == 0:
            raise ValueError(
                'Longitude variable not set and could not be ' 'discerned from the data.'
            )
        else:
            lon = lon[0]

    if alt is None:
        alt = [s for s in variables if 'altitude' in s]
        if len(alt) == 0:
            alt = [s for s in variables if 'alt' in s]
        if len(alt) == 0:
            raise ValueError(
                'Altitude variable not set and could not be ' 'discerned from the data.'
            )
        else:
            alt = alt[0]

    # Get additional variable names if none provided
    if azimuth is None:
        azimuth = [s for s in sorted(variables) if 'azimuth' in s][0]
        if len(azimuth) == 0:
            raise ValueError(
                'Azimuth variable not set and could not be ' 'discerned from the data.'
            )

    if elevation is None:
        elevation = [s for s in sorted(variables) if 'elevation' in s][0]
        if len(elevation) == 0:
            raise ValueError(
                'Elevation variable not set and could not be ' 'discerned from the data.'
            )

    if range_var is None:
        range_var = [s for s in sorted(variables) if 'range' in s][0]
        if len(range_var) == 0:
            raise ValueError('Range variable not set and could not be ' 'discerned from the data.')

    # Calculate the sweep indices if not passed in
    if sweep_start is None and sweep_end is None:
        az_diff = np.abs(np.diff(ds[azimuth].values))
        az_idx = az_diff > sweep_az_thresh

        el_diff = np.abs(np.diff(ds[elevation].values))
        el_idx = el_diff > sweep_el_thresh

        # Create index list
        az_index = list(np.where(az_idx)[0] + 1)
        el_index = list(np.where(el_idx)[0] + 1)
        index = sorted(az_index + el_index)

        index.insert(0, 0)
        index += [ds.sizes['time']]

        sweep_start_index = []
        sweep_end_index = []
        for i in range(len(index) - 1):
            sweep_start_index.append(index[i])
            sweep_end_index.append(index[i + 1] - 1)
            swp[index[i] : index[i + 1]] = i
    else:
        sweep_start_index = ds[sweep_start].values
        sweep_end_index = ds[sweep_end].values
        if sweep is None:
            for i in range(len(sweep_start_index)):
                swp[sweep_start_index[i] : sweep_end_index[i]] = i

    radar = pyart.testing.make_empty_ppi_radar(ds.sizes[range_var], ds.sizes['time'], nsweeps)

    radar.time['data'] = np.array(ds['time'].values)

    # Add lat, lon, and alt
    radar.latitude['data'] = np.array(ds[lat].values)
    radar.longitude['data'] = np.array(ds[lon].values)
    radar.altitude['data'] = np.array(ds[alt].values)

    # Add sweep information
    radar.sweep_number['data'] = swp
    radar.sweep_start_ray_index['data'] = sweep_start_index
    radar.sweep_end_ray_index['data'] = sweep_end_index
    radar.sweep_mode['data'] = np.array(sweep_mode)
    radar.scan_type = sweep_mode

    # Add elevation, azimuth, etc...
    radar.azimuth['data'] = np.array(ds[azimuth])
    radar.elevation['data'] = np.array(ds[elevation])
    radar.fixed_angle['data'] = np.array(ds[elevation].values[0])
    radar.range['data'] = np.array(ds[range_var].values)

    # Calculate radar points in lat/lon
    radar.init_gate_altitude()
    radar.init_gate_longitude_latitude()

    # Add the fields to the radar object
    fields = {}
    for v in variables:
        ref_dict = pyart.config.get_metadata(v)
        ref_dict['data'] = np.array(ds[v].values)
        fields[v] = ref_dict
    radar.fields = fields

    return radar


def convert_to_potential_temp(
    ds=None,
    temp_var_name=None,
    press_var_name=None,
    temperature=None,
    pressure=None,
    temp_var_units=None,
    press_var_units=None,
):
    """
    Converts temperature to potential temperature.

    Parameters
    ----------
    ds : xarray.DataSet
        ACT Xarray dataset
    temp_var_name : str
        Temperature variable name in the ACT Xarray dataset containing
        temperature data to convert.
    press_var_name : str
        Pressure variable name in the ACT Xarray dataset containing the
        pressure data to use in conversion. If not set or set to None will
        use values from pressure keyword.
    pressure : int, float, numpy array
        Optional pressure values to use instead of using values from xarray
        dataset. If set must also set press_var_units keyword.
    temp_var_units : string
        Pint recognized units string for temperature data. If set to None will
        use the units attribute under temperature variable in ds.
    press_var_units : string
        Pint recognized units string for pressure data. If set to None will
        use the units attribute under pressure variable in the dataset. If using
        the pressure keyword this must be set.

    Returns
    -------
    potential_temperature : None, int, float, numpy array
        The converted temperature to potential temperature or None if something
        goes wrong.

    References
    ----------
    May, R. M., Arms, S. C., Marsh, P., Bruning, E., Leeman, J. R., Goebbert,
    K., Thielen, J. E., and Bruick, Z., 2021: MetPy: A Python Package for
    Meteorological Data. Unidata, https://github.com/Unidata/MetPy,
    doi:10.5065/D6WW7G29.

    """
    potential_temp = None
    if temp_var_units is None and temp_var_name is not None:
        temp_var_units = ds[temp_var_name].attrs['units']
    if press_var_units is None and press_var_name is not None:
        press_var_units = ds[press_var_name].attrs['units']

    if press_var_units is None:
        raise ValueError(
            "Need to provide 'press_var_units' keyword " "when using 'pressure' keyword"
        )
    if temp_var_units is None:
        raise ValueError(
            "Need to provide 'temp_var_units' keyword " "when using 'temperature' keyword"
        )

    if temperature is not None:
        temperature = metpy.units.units.Quantity(temperature, temp_var_units)
    else:
        temperature = metpy.units.units.Quantity(ds[temp_var_name].values, temp_var_units)

    if pressure is not None:
        pressure = metpy.units.units.Quantity(pressure, press_var_units)
    else:
        pressure = metpy.units.units.Quantity(ds[press_var_name].values, press_var_units)

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=RuntimeWarning)
        potential_temp = metpy.calc.potential_temperature(pressure, temperature)
    potential_temp = potential_temp.to(temp_var_units).magnitude

    return potential_temp


def height_adjusted_temperature(
    ds=None,
    temp_var_name=None,
    height_difference=0,
    height_units='m',
    press_var_name=None,
    temperature=None,
    temp_var_units=None,
    pressure=101.325,
    press_var_units='kPa',
):
    """
    Converts temperature for change in height.

    Parameters
    ----------
    ds : xarray.DataSet, None
        Optional Xarray dataset for retrieving pressure and temperature values.
        Not needed if using temperature keyword.
    temp_var_name : str, None
        Optional temperature variable name in the Xarray dataset containing the
        temperature data to use in conversion. If not set or set to None will
        use values from temperature keyword.
    height_difference : int, float
        Required difference in height to adjust pressure values. Positive
        values to increase height negative values to decrease height.
    height_units : str
        Units of height value.
    press_var_name : str, None
        Optional pressure variable name in the Xarray dataset containing the
        pressure data to use in conversion. If not set or set to None will
        use values from pressure keyword.
    temperature : int, float, numpy array, None
        Optional temperature values to use instead of values in the dataset.
    temp_var_units : str, None
        Pint recognized units string for temperature data. If set to None will
        use the units attribute under temperature variable in the dataset.
        If using the temperature keyword this must be set.
    pressure : int, float, numpy array, None
        Optional pressure values to use instead of values in the dataset.
        Default value of sea level pressure is set for ease of use.
    press_var_units : str, None
        Pint recognized units string for pressure data. If set to None will
        use the units attribute under pressure variable in the dataset.
        If using the pressure keyword this must be set. Default value of
        sea level pressure is set for ease of use.

    Returns
    -------
    adjusted_temperature : None, int, float, numpy array
        The height adjusted temperature or None if something goes wrong.

    References
    ----------
    May, R. M., Arms, S. C., Marsh, P., Bruning, E., Leeman, J. R., Goebbert,
    K., Thielen, J. E., and Bruick, Z., 2021: MetPy: A Python Package for
    Meteorological Data. Unidata, https://github.com/Unidata/MetPy,
    doi:10.5065/D6WW7G29.

    """
    adjusted_temperature = None
    if temp_var_units is None and temperature is None:
        temp_var_units = ds[temp_var_name].attrs['units']
    if temp_var_units is None:
        raise ValueError(
            "Need to provide 'temp_var_units' keyword when " 'providing temperature keyword values.'
        )

    if temperature is not None:
        temperature = metpy.units.units.Quantity(temperature, temp_var_units)
    else:
        temperature = metpy.units.units.Quantity(ds[temp_var_name].values, temp_var_units)

    if press_var_name is not None:
        pressure = metpy.units.units.Quantity(ds[press_var_name].values, press_var_units)
    else:
        pressure = metpy.units.units.Quantity(pressure, press_var_units)

    adjusted_pressure = height_adjusted_pressure(
        height_difference=height_difference,
        height_units=height_units,
        pressure=pressure.magnitude,
        press_var_units=press_var_units,
    )
    adjusted_pressure = metpy.units.units.Quantity(adjusted_pressure, press_var_units)

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=RuntimeWarning)
        adjusted_temperature = metpy.calc.dry_lapse(adjusted_pressure, temperature, pressure)
    adjusted_temperature = adjusted_temperature.to(temp_var_units).magnitude

    return adjusted_temperature


def height_adjusted_pressure(
    ds=None,
    press_var_name=None,
    height_difference=0,
    height_units='m',
    pressure=None,
    press_var_units=None,
):
    """
    Converts pressure for change in height.

    Parameters
    ----------
    ds : xarray.DataSet, None
        Optional Xarray dataset for retrieving pressure values. Not needed if
        using pressure keyword.
    press_var_name : str, None
        Optional pressure variable name in the Xarray dataset containing the
        pressure data to use in conversion. If not set or set to None will
        use values from pressure keyword.
    height_difference : int, float
        Required difference in height to adjust pressure values. Positive
        values to increase height negative values to decrease height.
    height_units : str
        Units of height value.
    pressure : int, float, numpy array, None
        Optional pressure values to use instead of values in the dataset.
    press_var_units : str, None
        Pint recognized units string for pressure data. If set to None will
        use the units attribute under pressure variable in the dataset.
        If using the pressure keyword this must be set.

    Returns
    -------
    adjusted_pressure : None, int, float, numpy array
        The height adjusted pressure or None if something goes wrong.

    References
    ----------
    May, R. M., Arms, S. C., Marsh, P., Bruning, E., Leeman, J. R., Goebbert,
    K., Thielen, J. E., and Bruick, Z., 2021: MetPy: A Python Package for
    Meteorological Data. Unidata, https://github.com/Unidata/MetPy,
    doi:10.5065/D6WW7G29.

    """
    adjusted_pressure = None
    if press_var_units is None and pressure is None:
        press_var_units = ds[press_var_name].attrs['units']

    if press_var_units is None:
        raise ValueError(
            "Need to provide 'press_var_units' keyword when " 'providing pressure keyword values.'
        )

    if pressure is not None:
        pressure = metpy.units.units.Quantity(pressure, press_var_units)
    else:
        pressure = metpy.units.units.Quantity(ds[press_var_name].values, press_var_units)

    height_difference = metpy.units.units.Quantity(height_difference, height_units)

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=RuntimeWarning)
        adjusted_pressure = metpy.calc.add_height_to_pressure(pressure, height_difference)
    adjusted_pressure = adjusted_pressure.to(press_var_units).magnitude

    return adjusted_pressure


def arm_site_location_search(site_code='sgp', facility_code=None):
    """
    Parameters
    ----------
    site_code : str
        ARM site code to retrieve facilities and coordinate information. Example and default
        is 'sgp'.
    facility_code : str or None
        Facility code or codes for the ARM site provided. If None is provided, all facilities are returned.
        Example string for multiple facilities is 'A4,I5'.

    Returns
    -------
    coord_dict : dict
        A dictionary containing the facility chosen coordinate information or all facilities
        if None for facility_code and their respective coordinates.

    """
    headers = {
        'Content-Type': 'application/json',
        'user-agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/67.0.3396.99 Safari/537.36',
    }
    # Return all facilities if facility_code is None else set the query to include
    # facility search
    if facility_code is None:
        query = "site_code:" + site_code
    else:
        query = "site_code:" + site_code + " AND facility_code:" + facility_code

    # Search aggregation for elastic search
    json_data = {
        "aggs": {
            "distinct_facility_code": {
                "terms": {
                    "field": "facility_code.keyword",
                    "order": {"_key": "asc"},
                    "size": 7000,
                },
                "aggs": {
                    "hits": {
                        "top_hits": {
                            "_source": [
                                "site_type",
                                "site_code",
                                "facility_code",
                                "location",
                            ],
                            "size": 1,
                        },
                    },
                },
            },
        },
        "size": 0,
        "query": {
            "query_string": {
                "query": query,
            },
        },
    }

    # Uses requests to grab metadata from adc.arm.gov.
    response = requests.post(
        'https://adc.arm.gov/elastic/metadata/_search', headers=headers, json=json_data
    )
    # Loads the text to a dictionary
    response_dict = json.loads(response.text)

    # Searches dictionary for the site, facility and coordinate information.
    coord_dict = {}
    # Loop through each facility.
    for i in range(len(response_dict['aggregations']['distinct_facility_code']['buckets'])):
        site_info = response_dict['aggregations']['distinct_facility_code']['buckets'][i]['hits'][
            'hits'
        ]['hits'][0]['_source']
        site = site_info['site_code']
        facility = site_info['facility_code']
        # Some sites do not contain coordinate information, return None if that is the case.
        if site_info['location'] is None:
            coords = {'latitude': None, 'longitude': None}
        else:
            lat, lon = site_info['location'].split(',')
            lat = float(lat)
            lon = float(lon)
            coords = {'latitude': lat, 'longitude': lon}
        coord_dict.setdefault(site + ' ' + facility, coords)

    return coord_dict


def calculate_percentages(ds, fields, time=None, time_slice=None, threshold=None, fill_value=0.0):
    """
    This function calculates percentages of different fields of a dataset.

    Parameters
    ----------
    ds : ACT Dataset
        The ACT dataset to calculate the percentages on.
    fields : list
        A list of all the fields to use in the percentage calculations.
    time : datetime
        A single datetime to calculate percentages on if desired. Default
        is None and all data will be included.
    time_slice : tuple
        A tuple of two datetimes to grab all data between those two datatimes.
        Default is None and all data will be included.
    threshold : float
        Threshold in which anything below will be considered invalid.
        Default is None.
    fill_value : float
        Fill value for invalid data. Only used if a threshold is provided.

    Returns
    -------
    percentages : dict
        A dictionary containing the fields provided and their corresponding
        percentage that was calculated.

    """
    # Copy Dataset so we are not overriding the data.
    ds_percent = ds.copy()

    # Check if any incorrect values based on a threshold and replace with a fill
    # value.
    if threshold is not None:
        for field in fields:
            ds_percent[field] = ds_percent[field].where(ds_percent[field] > threshold, fill_value)

    # Raise warning if negative values present in a field.
    if threshold is None:
        for field in fields:
            res = np.all(ds_percent[field].values >= 0.0)
            if not res:
                warnings.warn(
                    f"{field} contains negatives values, consider using a threshold.",
                    UserWarning,
                )

    # Select the data based on time, multiple times within a slice, or
    # a sample of times per a timestep.
    if time is not None:
        ds_percent = ds_percent.sel(time=time)
    elif time_slice is not None:
        ds_percent = ds_percent.sel(time=slice(time_slice[0], time_slice[1]))
    else:
        warnings.warn(
            "No time parameter used, calculating a mean for each field for the whole dataset.",
            UserWarning,
        )

    # Calculate concentration percentage of each field in the air.
    values = [ds_percent[field].mean(skipna=True).values for field in fields]
    total = sum(values)
    percent_values = [(value / total) * 100 for value in values]

    # Create a dictionary of the fields and their percentages.
    percentages = {}
    for i, j in zip(fields, percent_values):
        percentages[i] = j
    ds_percent.close()
    return percentages


def convert_2d_to_1d(
    ds,
    parse=None,
    variables=None,
    keep_name_if_one=False,
    use_dim_value_in_name=False,
    dim_labels=None,
):
    """
    Function to  convert a single 2D variable into multiple 1D
    variables using the second dimension in the new variable name.

    Parameters
    ----------
    ds: xarray.dataset
        Object containing 2D variable to be converted
    parse: str or None
        Coordinate dimension name to parse along. If set to None will
        guess the non-time dimension is the parse dimension.
    variables: str or list of str
        Variable name or names to parse. If not provided will attempt to
        parse all two dimensional variables with the parse coordinate
        dimension.
    keep_name_if_one: boolean
        Option to not modify the variable name if the coordinate dimension
        has only one value. Essentially converting a 2D (i.e. (100,1)
        variable into a 1D variable (i.e. (100)).
    use_dim_value_in_name: boolean
        Option to use value from the coordinate dimension in new variable
        name instead of indexing number. Will use the value prepended
        to the units of the dimension.
    dim_labels: str or list of str
        Allows for use of custom label to append to end of variable names

    Returns
    -------
        A new object copied from input object with the multi-dimensional
        variable split into multiple single-dimensional variables.

    Example
    -------
    # This will get the name of the coordinate dimension so it does not need to
    # be hard coded.
    >>> parse_dim = (list(set(list(ds.dims)) - set(['time'])))[0]

    # Now use the parse_dim name to parse the variable and return new object.
    >>> new_ds = convert_2d_to_1d(ds, parse=parse_dim)

    """
    # If no parse dimension name given assume it is the one not equal to 'time'
    if parse is None:
        parse = (list(set(list(ds.dims)) - {'time'}))[0]

    new_ds = ds.copy()

    if variables is not None and isinstance(variables, str):
        variables = [variables]

    if variables is None:
        variables = list(new_ds.variables)

    if dim_labels is not None and isinstance(dim_labels, (str,)):
        dim_labels = [dim_labels]

    # Check if we want to keep the names the same if the second dimension
    # is of size one.
    num_dims = 1
    if keep_name_if_one:
        num_dims = 2

    parse_values = ds[parse].values
    for var in variables:
        if var == parse:
            continue
        # Check if the parse dimension is in the dimension tuple
        if parse in new_ds[var].dims:
            if len(new_ds[parse]) >= num_dims:
                for i in range(0, new_ds.sizes[parse]):
                    if dim_labels is not None:
                        new_var_name = '_'.join([var, dim_labels[i]])
                    elif use_dim_value_in_name:
                        level = str(parse_values[i]) + ds[parse].attrs['units']
                        new_var_name = '_'.join([var, parse, level])
                    else:
                        new_var_name = '_'.join([var, parse, str(i)])
                    new_var = new_ds[var].copy()
                    new_ds[new_var_name] = new_var.isel(indexers={parse: i})

                    try:
                        ancillary_variables = new_ds[new_var_name].attrs['ancillary_variables']
                        current_qc_var_name = ds.qcfilter.check_for_ancillary_qc(
                            var, add_if_missing=False
                        )
                        if current_qc_var_name is not None:
                            ancillary_variables = ancillary_variables.replace(
                                current_qc_var_name, 'qc_' + new_var_name
                            )
                            new_ds[new_var_name].attrs['ancillary_variables'] = ancillary_variables
                    except KeyError:
                        pass

                # Remove the old 2D variable after extracting
                del new_ds[var]

            else:
                # Keep the same name but remove the dimension equal to size 1
                new_ds[var] = new_ds[var].squeeze(dim=parse)

    return new_ds



"""
Time-Height Scatter Plot
------------------------
This will show how to use the time-height scatter
plot function that's part of the TimeSeries Display.

"""

import matplotlib.pyplot as plt

import act
from act.tests import sample_files

# Read in radiosonde data
ds = act.io.arm.read_arm_netcdf(sample_files.EXAMPLE_SONDE1)

# Create scatter plots of the sonde data
display = act.plotting.TimeSeriesDisplay(ds, figsize=(7, 6), subplot_shape=(2,))
display.time_height_scatter('tdry', plot_alt_field=True, subplot_index=(0,))
display.time_height_scatter('rh', subplot_index=(1,), cvd_friendly=True, day_night_background=True)
plt.tight_layout()
ds.close()

plt.show()



"""
Plotting QC Flags
-----------------

Simple example for cleaning up a dataset and
plotting the data and its QC flags

Author: Adam Theisen
"""

from arm_test_data import DATASETS
from matplotlib import pyplot as plt

import act

# Read in sample MET data
filename_met = DATASETS.fetch('sgpmetE13.b1.20190101.000000.cdf')
ds = act.io.arm.read_arm_netcdf(filename_met)

# In order to utilize all the ACT QC modules and plot the QC,
# we need to clean up the dataset to follow CF standards
ds.clean.cleanup()


# Plot data
# Creat Plot Display
variable = 'temp_mean'
display = act.plotting.TimeSeriesDisplay(ds, figsize=(15, 10), subplot_shape=(2,))

# Plot temperature data in top plot
display.plot(variable, subplot_index=(0,))

# Plot QC data
display.qc_flag_block_plot(variable, subplot_index=(1,))
plt.show()
